"""Meta classes for OrientDb."""
import abc
import copy
import json
import logging
import os
import random
import socket
import time
from abc import abstractmethod
from collections import OrderedDict, defaultdict
from http.client import RemoteDisconnected
from shutil import copyfileobj
from types import GeneratorType
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union
from urllib.request import Request, urlopen

import numpy as np
import pandas as pd
import requests
import sqlalchemy as sqla
import xmltodict
from pyorientdb import OrientDB, orient
from pyorientdb.exceptions import (PyOrientCommandException,
                                   PyOrientIndexException,
                                   PyOrientSecurityAccessException)
from pyorientdb.otypes import OrientRecord
from sqlalchemy.sql.schema import Table
from sqlalchemy_utils import create_database, database_exists
from tqdm import tqdm

import ebel.database
from ebel.cache import set_mysql_interactive
from ebel.config import get_config_as_dict, get_config_value, write_to_config
from ebel.constants import DEFAULT_ODB, RID
from ebel.manager.orientdb import urls as default_urls
from ebel.manager.orientdb.odb_structure import (Edge, Generic, Node, OClass,
                                                 OIndex, OProperty)
from ebel.tools import BelRdb, chunks, get_file_path, get_standard_name

type_map_inverse = {v: k for k, v in orient.type_map.items()}

logger = logging.getLogger(__name__)


class ExceptionClassNotExists(Exception):
    """Package specific exception class definition."""

    def __init__(self, message):
        """Init method."""
        self.message = message


class Graph(abc.ABC):
    """Generic parent class for BioDBs."""

    def __init__(
        self,
        generics: Tuple[Generic] = (),
        nodes: Tuple[Node] = (),
        edges: Tuple[Edge] = (),
        indices: Tuple[OIndex] = (),
        urls: dict = None,
        biodb_name: str = "",
        tables_base=None,
        config_params: Optional[dict] = None,
        overwrite_config: bool = False,
    ):
        """Init method."""
        self.generic_classes: Tuple[OClass] = generics
        self.node_classes: Tuple[OClass] = nodes
        self.edge_classes: Tuple[OClass] = edges
        self.class_indices: Tuple[OIndex] = indices
        self.tables_base = tables_base  # sqlalchemy base

        self.urls = urls if urls else {}
        self.biodb_name = biodb_name

        self.odb_db_name = config_params["db"] if config_params and "db" in config_params else None
        self.odb_user = config_params["user"] if config_params and "user" in config_params else None
        self.odb_password = config_params["password"] if config_params and "password" in config_params else None
        self.odb_server = config_params["server"] if config_params and "server" in config_params else "localhost"
        self.odb_port = config_params["port"] if config_params and "port" in config_params else "2424"
        self.odb_user_reader = None
        self.odb_user_reader_password = None
        # Root password should not be set, but can be
        self._odb_root_password = None

        self.__config_params_check(overwrite_config)
        self.client: OrientDB = self.get_client()

        self.create_all_classes()
        self.cluster_ids = {}

        rdb = BelRdb()
        self.engine = rdb.engine
        self.session = rdb.session

        if not (get_config_value("DATABASE", "sqlalchemy_connection_string") or database_exists(self.engine.url)):
            if str(self.engine.url).startswith("mysql"):
                set_mysql_interactive()

            else:
                create_database(self.engine.url)

    def __config_params_check(self, overwrite_config: bool = False):
        """Go through passed/available configuration params."""
        # Set the client
        config_dict = get_config_as_dict()

        credentials = {
            "name": self.odb_db_name,
            "user": self.odb_user,
            "password": self.odb_password,
            "server": self.odb_server,
            "port": int(self.odb_port),
            "root_password": self._odb_root_password,
        }

        # If values are passed
        if all(credentials.values()):
            #  If there is no "DEFAULT_ODB" section in the configuration file, then write these values to it
            if DEFAULT_ODB not in config_dict:
                for option, value in credentials.items():
                    if "root" not in option and overwrite_config:  # Don't want to write root to config file
                        write_to_config(section=DEFAULT_ODB, option=option, value=value)

            self.set_configuration_parameters()

        # If there is a "DEFAULT_ODB" section in the configuration file, then read these values
        elif DEFAULT_ODB in config_dict:
            self.set_configuration_parameters()

        # No parameters passed and no saved parameters in config means user needs to provide some info
        else:
            missing_params = ", ".join({key for key, val in credentials.items() if val is None})
            logger.info(f"Initial configuration parameters missing. Missing parameters: {missing_params}")
            raise ValueError(f"Please provide initial configuration parameters. Missing parameters: {missing_params}")

    def execute(self, command_str: str) -> List[OrientRecord]:
        """Execute a command directly in the OrientDB server.

        Parameters
        ----------
        command_str: str
            The SQL to be executed

        Raises
        ------
        PyOrientCommandException
            Caused by a disconnect to the ODB server. eBEL will try to reconnect if possible.

        Returns
        -------
        ODB response.
        """
        try:
            return self.client.command(command_str)

        # TODO: following exceptions seems not to cover connection error
        # except (PyOrientCommandException, PyOrientSecurityAccessException):
        except:
            # Try to reconnect
            self.client.close()
            self.client = self.get_client()
            # self.client.db_open(self.odb_name, self.odb_user, self.odb_password)
            # print(command_str)
            return self.client.command(command_str)

    def set_configuration_parameters(self):
        """Set configuration for OrientDB database client instance using configuration file or passed params."""
        odb_config = get_config_as_dict().get(DEFAULT_ODB)

        self.odb_db_name = self.odb_db_name or odb_config.get("name")
        self.odb_user = self.odb_user or odb_config.get("user")
        self.odb_password = self.odb_password or odb_config.get("password")
        self.odb_server = self.odb_server or odb_config.get("server")
        self.odb_port = int(self.odb_port or odb_config.get("port") or "2424")
        self.odb_user_reader = self.odb_user_reader or odb_config.get("user_reader") or None
        self.odb_user_reader_password = self.odb_user_reader_password or odb_config.get("user_reader_password") or None

        # Root password should not be written in the config file, but it's possible
        self._odb_root_password = self._odb_root_password or odb_config.get("root_password")

    def get_client(self) -> OrientDB:
        """Attempts to connect to the OrientDB client. This is currently done by using session tokens."""
        client = ebel.database.get_orientdb_client(
            self.odb_server,
            self.odb_port,
            self.odb_db_name,
            self.odb_user,
            self.odb_password,
            self._odb_root_password,
            self.odb_user_reader,
            self.odb_user_reader_password,
        )

        return client

    def __repr__(self):
        """Represent the class."""
        template = "{{BioDatabase:{class_name}}}[url:{url}, edges:{edges}, generics:{generics}]"
        representation = template.format(
            class_name=self.__class__.__name__,
            url=self.urls,
            edges={k: v for k, v in self.number_of_edges.items() if v},
            nodes={k: v for k, v in self.number_of_nodes.items() if v},
            generics={k: v for k, v in self.number_of_generics.items() if v},
        )
        return representation

    @abstractmethod
    def insert_data(self) -> Dict[str, int]:
        """Insert all generic data."""

    @abstractmethod
    def update_interactions(self) -> int:
        """Insert all generic data."""

    def clear_and_import_data(self) -> Dict[str, int]:
        """Clears the associated table and inserts the data from raw downloaded data."""
        inserted = {}
        logger.info(f"Clear generics for {self.biodb_name}")
        if self.generic_classes:
            self.clear_generics()

        if self.tables_base:
            self.recreate_tables()

        try:
            inserted = self.insert_data()
            logger.info(f"Successfully inserted data for {self.biodb_name}")

        except FileNotFoundError:
            logging.error(f"Data file for {self.biodb_name.upper()} was not found. Skipping...")

        except Exception as e:
            logger.error(f"Failed to insert data for {self.biodb_name}!", exc_info=e)

        return inserted

    def create_index_rdbms(self, table_name: str, columns):
        """Creates index on column(s) in RDBMS."""
        if isinstance(columns, str):
            columns = [columns]
        sql_columns = ",".join(columns)
        index_name = f"idx_{table_name}_" + "_".join(columns)
        self.engine.execute(f"CREATE INDEX {index_name} ON {table_name} ({sql_columns})")

    def clear_edges_by_bel_doc_rid(self, bel_document_rid: str, even_if_other_doc_rids_exists=True):
        """Delete all edges linked to a specified BEL document rID."""
        changes = 0
        if even_if_other_doc_rids_exists:
            sql = f"DELETE FROM `bel_relation` WHERE {bel_document_rid} IN document"
            changes = self.execute(sql)[0]

        else:
            sql = (
                f"SELECT @rid.asString() AS rid, document.asString() FROM `bel_relation` "
                f"WHERE {bel_document_rid} IN document"
            )
            for r in self.query_get_dict(sql):
                new_rids = [x.strip() for x in r["document"][1:-1].split(",")]
                new_rids.remove(bel_document_rid)
                if new_rids:
                    new_rids_str = ",".join(new_rids)
                    self.execute(f"UPDATE {r['rid']} SET document = [{new_rids_str}]")
                    changes += 1
                else:
                    self.execute(f"DELETE EDGE {r['rid']}")
                    changes += 1

        self.execute(f"DELETE from bel_document where @rid = {bel_document_rid}")

        return changes

    def clear_documents(self) -> int:
        """Clear all document info. Returns number of deleted documents."""
        return self.execute("Delete from `bel_document`")[0]

    def get_number_of_bel_statements_by_document_rid(self, bel_document_rid: str) -> int:
        """Return BEL statement count with a given document ID."""
        sql = f"Select count(*) as num from bel_relation where  document = {bel_document_rid}"
        return self.execute(sql)[0].oRecordData["num"]

    def get_documents(self):
        """Return all document info as pandas DataFrame."""
        return self.query("Select * from bel_document")

    def get_documents_as_dict(self):
        """Return all document info as pandas DataFrame."""
        return [x.oRecordData for x in self.execute("Select * from bel_document")]

    def add_keyword(self, keyword: str, description: str) -> pd.DataFrame:
        """Add a keyword and description used to tagging BEL documents.

        Parameters
        ----------
        keyword : str
            The name of a project the work is based on or type of work
        description : str
            Detailed explanation of the keyword
        """
        sql = f"INSERT INTO keyword CONTENT {'label': {keyword}, 'description': {description}}"
        return self.query(sql)

    def get_info_class(self, class_name):
        """Return info about class."""
        sql = "Select * from (SELECT expand( classes ) FROM metadata:schema) where name='{}' limit 1"
        return self.execute(sql.format(class_name))[0].oRecordData

    def get_info_properties(self, class_name: str, short: bool = True):
        """Get the property information for a specified table."""
        sql_temp = "select expand(properties) from (select expand(classes) from metadata:schema) where name = '{}'"
        sql = sql_temp.format(class_name)
        o_record_datas = [x.oRecordData for x in self.execute(sql)]

        if short:
            props = [{"name": x["name"], "type": type_map_inverse[x["type"]]} for x in o_record_datas]

        else:
            props = o_record_datas

        return props

    def entry_exists(self, class_name, **params) -> bool:
        """Check if class_name whith parameters exists."""
        where_list = []
        for column, value in params.items():
            if isinstance(value, (str, list, dict)):
                where_list.append("`{}` = {}".format(column, json.dumps(value)))
            elif isinstance(value, (int, float)):
                where_list.append("`{}` = {}".format(column, value))

        where = " and ".join(where_list)
        sql = "Select 1 from `{}` where {} limit 1".format(class_name, where)
        results = self.execute(sql)

        return bool(len(results))

    def query(self, sql: str) -> pd.DataFrame:
        """Return a pandas DataFrame results table."""
        results = self.execute(sql)
        return pd.DataFrame([x.oRecordData for x in results])

    def query_get_dict(self, sql: str) -> List[dict]:
        """Return list of dictionaries using a given SQL query."""
        results = self.execute(sql)
        return [x.oRecordData for x in results]

    def query_class(
        self,
        class_name: str,
        limit: int = 0,
        skip: int = 0,
        columns: Iterable[str] = None,
        with_rid=True,
        with_class=False,
        print_sql: bool = False,
        group_by: List[str] = None,
        distinct=False,
        as_dataframe: bool = False,
        where_list: Tuple[str] = (),
        **params,
    ) -> Union[List[dict], pd.DataFrame]:
        """Query class by params and returns list of pyorient.OrientRecord."""
        if not self.class_exists(class_name):
            raise ExceptionClassNotExists("Class {} not exists in database.".format(class_name))

        where = self.__get_sql_where_part(params, where_list)

        sql_limit = ""
        if limit:
            sql_limit = "LIMIT {}".format(limit)

        sql_skip = ""
        if skip:
            sql_skip = "SKIP {}".format(skip)

        if isinstance(columns, Iterable):
            cols = columns

        else:
            cols = ["*"]

        if with_rid:
            cols.append("@rid.asString()")

        if with_class:
            cols.append("@class.asString()")

        sql_cols = ", ".join(cols)

        if isinstance(group_by, Iterable):
            group_by = "GROUP BY " + ", ".join(group_by)

        else:
            group_by = ""

        if distinct and len(cols) == 1:
            sql_cols = "distinct({})".format(sql_cols)

        sql_temp = "SELECT {sql_cols} FROM `{class_name}` {where} {group_by} {sql_limit} {sql_skip}"

        sql = sql_temp.format(
            sql_cols=sql_cols,
            class_name=class_name,
            where=where,
            sql_limit=sql_limit,
            sql_skip=sql_skip,
            group_by=group_by,
        )

        if print_sql:
            print(sql)

        results = self.execute(sql)

        if as_dataframe:
            return_value = pd.DataFrame([x.oRecordData for x in results])

        else:
            return_value = [x.oRecordData for x in results]

        return return_value

    def query_class_chunks(
        self,
        class_name: str,
        chunk_size: int = 10000,
        columns: Iterable[str] = None,
        with_rid=True,
        with_class=False,
        print_sql: bool = False,
        group_by: List[str] = None,
        distinct: bool = False,
    ):
        """Query class by params and only return a set of results in batches. Creates a generator."""
        number_entries = self.get_number_of_class(class_name)
        total_num_chunks = (number_entries // chunk_size) + 1

        chunk_index = 0
        while chunk_index < total_num_chunks:
            generic_table_chunk = self.query_class(
                class_name=class_name,
                limit=chunk_size,
                skip=chunk_size * chunk_index,
                with_rid=with_rid,
                with_class=with_class,
                print_sql=print_sql,
                group_by=group_by,
                distinct=distinct,
                columns=columns,
            )
            yield generic_table_chunk
            chunk_index += 1

    def query_rid(self, rid, columns: list = None):
        """Query specified columns of a given rID entry."""
        table_columns = columns if columns else []
        columns_in_sql = ", ".join(table_columns)
        results = self.execute(f"Select {columns_in_sql} from {rid}")
        if results:
            return [x.oRecordData for x in results][0]

    def download(
        self,
        url_dict: Dict[str, str] = None,
        biodb: str = None,
        expiration_days: int = 100,
    ) -> Dict[str, bool]:
        """Download url to file_path if not older than expiration_days."""
        if not url_dict and hasattr(self, "urls"):
            url_dict = self.urls

        downloaded = {}

        for key, url in url_dict.items():
            downloaded[key] = self.download_file(url, self.biodb_name, expiration_days)

        return downloaded

    @staticmethod
    def download_file(url: str, biodb: str, expiration_days: int = 100, addtional_header: dict = None) -> bool:
        """Download file. Returns True if it was needed to download the file."""
        header = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/"
            "537.36 (KHTML, like Gecko) Chrome/50.0.2661.75 Safari/537.36",
            "X-Requested-With": "XMLHttpRequest",
        }
        if addtional_header:
            header.update(addtional_header)

        expiration_secs = expiration_days * 60 * 60 * 24
        now = time.time()
        downloaded = False
        file_path = get_file_path(url, biodb)
        file_exists = os.path.exists(file_path)

        if file_exists:
            expired = now - os.path.getmtime(file_path) > expiration_secs
        else:
            expired = True

        if not file_exists or expired:
            logger.info(f"Download {url}")
            try:
                with urlopen(Request(url, headers=header)) as response, open(file_path, "wb") as out_file:
                    copyfileobj(response, out_file)

                downloaded = True

            except RemoteDisconnected:
                logger.error(
                    f"{url} could not be downloaded! Website appears to be down.",
                    exc_info=False,
                )

            except Exception:
                logger.error(f"{url} could not be downloaded!", exc_info=True)

        return downloaded

    def index_exists(self, index_name: str):
        """Check if index_name exists."""
        sql = """Select 1 as exists from (select expand(indexes)
                  from metadata:indexmanager) where name = '{}'""".format(
            index_name
        )
        return len(self.execute(sql)) > 0

    def create_index(self, index: OIndex):
        """Create index."""
        sql = "CREATE INDEX {index_name} ON `{class_name}` ({columns}) {index_type}"

        index_name = self.get_index_name(index)
        if not self.index_exists(index_name):
            sql = sql.format(
                class_name=index.class_name,
                columns=",".join(index.columns),
                index_name=self.get_index_name(index),
                index_type=index.index_type.value,
            )
            try:
                self.execute(sql)

            except PyOrientIndexException:
                errmsg = f"Index failure: {self.get_index_name(index)} on {index.class_name}. Check ODB version."
                logger.error(errmsg)

        return self

    def get_index_name(self, index: OIndex):
        """Return index name."""
        return index.class_name + "." + "__".join(index.columns)

    def create_all_classes(self):
        """Create all classes."""
        class_names = self.create_classes(self.classes)

        self.create_indices(self.class_indices)
        return class_names

    def create_all_indices(self):
        """Create indices."""
        for index in self.class_indices:
            self.create_index(index)

    def create_indices(self, indices: Tuple[OIndex]):
        """Create indices."""
        for index in indices:
            self.create_index(index)

    def drop_all_indices(self):
        """Drop indices."""
        for index in self.class_indices:
            self.drop_index(index)

    def drop_indices(self, indices: List[OIndex]):
        """Drop indices."""
        for index in indices:
            self.drop_index(index)

    def drop_index(self, index: OIndex):
        """Drop index."""
        self.execute("DROP INDEX {}".format(self.get_index_name(index)))

    def create_node_classes(self):
        """Create node classes."""
        class_names = self.create_classes(self.node_classes)
        self.create_indices(self.class_indices)
        return class_names

    def create_edge_classes(self):
        """Create edge classes."""
        class_names = self.create_classes(self.edge_classes)
        self.create_indices(self.class_indices)
        return class_names

    def create_generic_classes(self):
        """Create generic classes."""
        class_names = self.create_classes(self.generic_classes)
        self.create_indices(self.class_indices)
        return class_names

    @property
    def classes(self) -> Tuple[OClass, ...]:
        """Return generic, node and edge classes as List[OClass]."""
        return self.generic_classes + self.node_classes + self.edge_classes

    def create_classes(self, oclasses: Tuple[OClass] = None):
        """Create classes.

        list of classes (odb_structure.OClass)
        OrientDB class v, e or g
        for vertex, edge or generic
        """
        created = []

        for oclass in oclasses:
            self.create_class(oclass)
            created.append(oclass.name)
        return created

    def create_class(self, oclass: OClass, print_sql=False):
        """Create class.

        OrientDB class v, e or g
        for vertex, edge or generic
        """
        sql_class = "CREATE CLASS {name} IF NOT EXISTS {extends} {abstract}"

        sql_dict = {"name": "`" + oclass.name + "`", "extends": "", "abstract": ""}

        if oclass.extends:
            sql_dict["extends"] = "EXTENDS " + ", ".join(["`" + x + "`" for x in oclass.extends])
        if oclass.abstract:
            sql_dict["abstract"] = "ABSTRACT"

        sql = sql_class.format(**sql_dict)
        if print_sql:
            print(sql)
        self.execute(sql)

        if oclass.props:
            for prop in oclass.props:
                self.create_class_property(class_name=oclass.name, prop=prop)

        if isinstance(oclass, Edge):
            sql_in_out = f"CREATE PROPERTY {oclass.name}.{{}}  IF NOT EXISTS LINK {{}}"
            if oclass.in_out[0]:
                self.execute(sql_in_out.format("in", oclass.in_out[0]))
            if oclass.in_out[1]:
                self.execute(sql_in_out.format("out", oclass.in_out[1]))

    def create_class_property(self, class_name: str, prop: OProperty, print_sql: bool = False):
        """Create OrientDB class property."""
        sql_prop = "CREATE PROPERTY `{name}`.`{prop_name}`\
        IF NOT EXISTS {dtype} {linked_class} {linked_type} {mandatory}"

        linked_class = prop.linked_class or ""
        linked_type = prop.linked_type.value if prop.linked_type else ""
        mandatory = "(MANDATORY TRUE)" if prop.mandatory else ""
        sql = sql_prop.format(
            name=class_name,
            prop_name=prop.prop_name,
            dtype=prop.data_type.value,
            linked_class=linked_class,
            linked_type=linked_type,
            mandatory=mandatory,
        )
        if print_sql:
            print(sql)
        self.execute(sql)

    def class_exists(self, class_name: str) -> bool:
        """Check if OrientDB class exists."""
        sql = "Select 1 as exists from (select expand(classes) from metadata:schema) where name = '{}'"
        r = self.execute(sql.format(class_name))
        return len(r) > 0

    def classes_exists(self, list_of_class_names):
        """Check if list of OrientDB classes exists."""
        for c in list_of_class_names:
            if not self.class_exists(c):
                return False
        return True

    def drop_class(self, class_name: str):
        """Drop the specified table."""
        self.execute("DROP CLASS `{}`".format(class_name))

    def __drop_classes(self, classes: Iterable[OClass]):
        """Delete the classes in opposite order."""
        dropped = []
        for oclass in classes[::-1]:
            if self.class_exists(oclass.name) and oclass.own_class:
                self.drop_class(oclass.name)
                dropped.append(oclass.name)
        return dropped

    def drop_all_classes(self):
        """Drop all classes."""
        return self.__drop_classes(self.classes)

    def drop_generic_classes(self):
        """Drop all generic classes."""
        return self.__drop_classes(self.generic_classes)

    def drop_node_classes(self):
        """Drop all node classes."""
        return self.__drop_classes(self.node_classes)

    def drop_edge_classes(self):
        """Drop all edge classes."""
        return self.__drop_classes(self.edge_classes)

    def clear(self):
        """Clear (delete entries) from all classes."""
        return {
            "nodes": self.clear_nodes(),
            "edges": self.clear_edges(),
            "generics": self.clear_generics(),
        }

    def is_abstract_class(self, class_name: str) -> bool:
        """Returns true if class is abstract."""
        sql = """select 1 from (
                    select expand(classes) from metadata:schema
                    ) where name = '{}' and abstract = true"""
        return bool(self.execute(sql.format(class_name)))

    def clear_edges(self) -> Dict[str, int]:
        """Delete all edges."""
        deleted = {}
        sql_temp = "DELETE EDGE `{}`"
        if hasattr(self, "edge_classes"):
            for oclass in self.edge_classes:
                if oclass.own_class and not oclass.abstract and self.class_exists(oclass.name):
                    sql = sql_temp.format(oclass.name)
                    deleted[oclass.name] = self.execute(sql)[0]
        return deleted

    def clear_class(self, class_name):
        """Delete all entries from class if exists."""
        sql = "DELETE FROM `{}`".format(class_name)
        deleted = self.execute(sql)[0]
        return deleted

    def clear_nodes(self):
        """Delete all nodes."""
        deleted = {}
        sql_temp = "DELETE VERTEX {}"
        for oclass in self.node_classes:
            if self.class_exists(oclass.name) and oclass.own_class:
                sql = sql_temp.format(oclass.name)
                deleted[oclass.name] = self.execute(sql)[0]
        return deleted

    def clear_nodes_with_no_edges(self):
        """Delete all nodes from a class with no edges."""
        sql_temp = "Delete vertex {} where bothE().size()=0"
        deleted = {}
        for oclass in self.node_classes:
            if self.class_exists(oclass.name) and oclass.own_class:
                sql = sql_temp.format(oclass.name)
                deleted[oclass.name] = self.execute(sql)[0]
        return deleted

    def clear_generics(self):
        """Delete all entries in generic tables."""
        deleted = {}
        for oclass in self.generic_classes:
            if self.class_exists(oclass.name):
                deleted[oclass.name] = self.clear_class(oclass.name)
        return deleted

    def recreate_tables(self):
        """Recreate SQLAlchemy tables in relational database."""
        self.tables_base.metadata.drop_all(self.engine)
        self.tables_base.metadata.create_all(self.engine)
        self.session.commit()

    def clear_nodes_and_edges(self):
        """Delete all nodes and edges of a specific biodb."""
        number_of_deleted_edges = self.clear_edges()
        number_of_deleted_nodes = self.clear_nodes()
        return {"nodes": number_of_deleted_nodes, "edges": number_of_deleted_edges}

    def clear_all_nodes_and_edges(self):
        """Delete all nodes and edges in the whole database."""
        edges_deleted = self.execute("Delete edge E")[0]
        nodes_deleted = self.execute("Delete vertex V")[0]
        return {"edges": edges_deleted, "nodes": nodes_deleted}

    def clear_exp_edges(self):
        """Delete all DEA experiment associated edges."""
        deleted = self.clear_class(class_name="ko_relation")
        return deleted

    def recreate(self):
        """Recreate OrientDB collection."""
        self.clear_edges()
        self.clear_nodes()
        self.drop_edge_classes()
        self.drop_node_classes()
        self.create_all_classes()

    def table_exists(self, table: Table):
        """Checks if the table exists in RDBMS."""
        return sqla.inspect(self.engine).has_table(table)

    @property
    def number_of_generics(self) -> Dict[str, int]:
        """Returns for number of entries in OrientDB classes and RDB tables. Tables have priority."""
        numbers = {}

        if self.tables_base:
            for table_name, table in self.tables_base.metadata.tables.items():
                if self.table_exists(table_name):
                    sql = f"Select count(*) from `{table_name}`"
                    numbers[table_name] = self.engine.execute(sql).fetchone()[0]
                else:
                    numbers[table_name] = 0
        elif self.generic_classes:
            for oclass in self.generic_classes:
                numbers[oclass.name] = self.get_number_of_class(oclass.name)

        return numbers

    @staticmethod
    def __get_sql_where_part(params, where_list: Tuple[str] = ()):
        """Return a ODB SQL where part by params."""
        where_list = list(where_list)
        for column, value in params.items():
            if isinstance(value, (str, list, dict)):
                if value == "notnull":
                    where_list.append("`{}` IS NOT NULL".format(column))
                else:
                    where_list.append("`{}` = {}".format(column, json.dumps(value)))
            elif isinstance(value, (int, float)):
                where_list.append("`{}` = {}".format(column, value))
            elif value is None:
                where_list.append("`{}` IS NULL".format(column))
        where = ""
        if where_list:
            where = " WHERE " + " AND ".join(where_list)
        return where

    def get_number_of_class(self, class_name, distinct_column_name: str = None, **params):
        """Return count of unique values for a given class_name and column name."""
        column = "*"
        if distinct_column_name:
            column = f"distinct({distinct_column_name})"
        where = self.__get_sql_where_part(params)
        sql = f"Select count(`{column}`) from `{class_name}`{where}"
        return self.execute(sql)[0].oRecordData["count"]

    @property
    def number_of_nodes(self):
        """Return node count."""
        numbers = {}
        for oclass in self.node_classes:
            if oclass.own_class:
                self.get_number_of_class(oclass.name)
                numbers[oclass.name] = self.get_number_of_class(oclass.name)
        return numbers

    @property
    def number_of_edges(self):
        """Return edge count."""
        numbers = {}
        for oclass in self.edge_classes:
            if oclass.own_class:
                numbers[oclass.name] = self.get_number_of_class(oclass.name)
        return numbers

    def get_cluster_ids(self, class_name: str) -> list:
        """Get all cluster ids by class name."""
        sql_cids = "Select clusterIds from (select expand(classes) from metadata:schema) where name = '{}'"
        if class_name not in self.cluster_ids:
            cids = self.execute(sql_cids.format(class_name))[0].oRecordData["clusterIds"]
            self.cluster_ids[class_name] = cids
        return self.cluster_ids[class_name]

    def insert_record(self, class_name: str, value_dict: dict, print_sql=False) -> Optional[str]:
        """Insert new entry in class with values from dictionary. Returns rid."""
        sql = "insert into `{}` content {}".format(class_name, json.dumps(value_dict))

        if print_sql:
            print(sql)

        try:
            r = self.execute(sql)[0]
            return r._OrientRecord__rid

        except (PyOrientCommandException, socket.timeout):
            logger.error(f"Unable to execute\n{sql}", exc_info=True)

    def create_record(self, class_name: str, value_dict: dict) -> Optional[str]:
        """Create record/ insert into class_name with content of value_dict."""
        cid = random.choice(self.get_cluster_ids(class_name))

        try:
            r = self.client.record_create(cid, {"@" + class_name: value_dict})._OrientRecord__rid

        except (PyOrientCommandException, socket.timeout):
            logger.warning("Standard insert in odb_meta.create_record did not work!")
            r = self.insert_record(class_name, value_dict)

        except Exception:
            logger.error("Failed to create record", exc_info=True)
            return None

        return r

    def update_record(self, class_name: str, value_dict: dict) -> str:
        """Update record with content of value_dict."""
        cid = random.choice(self.get_cluster_ids(class_name))
        r = self.client.record_update(cid, {"@" + class_name: value_dict})
        return r._OrientRecord__rid

    def edge_exists(self, class_name: str, from_rid: str, to_rid: str, value_dict: dict = {}) -> str:
        """Check if edge exists. Return rid if exists else None."""
        data = copy.deepcopy(value_dict)  # deep copy and DO NOT change the dictionary!!!
        data.update({"out.@rid": from_rid, "in.@rid": to_rid})
        result = self.query_class(class_name, limit=1, **data)
        if result:
            return result[0][RID]

    def node_exists(
        self,
        class_name: str,
        value_dict: dict = {},
        check_for: Union[Iterable[str], str] = None,
        print_sql: bool = False,
    ) -> str:
        """Check if node exists. Return rid if exists else None."""
        check_for_dict = value_dict.copy()
        if check_for:
            check_for = [check_for] if isinstance(check_for, str) else check_for
            check_for_dict = {k: v for k, v in check_for_dict.items() if k in check_for}
        result = self.query_class(class_name=class_name, limit=1, print_sql=print_sql, **check_for_dict)
        if result:
            return result[0][RID]

    def create_edge(
        self,
        class_name: str,
        from_rid: str,
        to_rid: str,
        value_dict: dict = {},
        print_sql=False,
        if_not_exists=False,
        ignore_empty_values=False,
    ) -> str:
        """Create edge from from_rid(@rid) to to_rid(@rid) with content of value_dict."""
        content = ""
        if if_not_exists:
            edge_rid = self.edge_exists(class_name, from_rid, to_rid, value_dict)
            if edge_rid:
                return edge_rid

        if value_dict:
            if ignore_empty_values:
                value_dict = {k: v for k, v in value_dict.items()}
            content = "CONTENT {}".format(json.dumps(value_dict))
        sql_temp = "CREATE EDGE `{class_name}` FROM {from_rid} TO {to_rid} {content}"
        sql = sql_temp.format(class_name=class_name, from_rid=from_rid, to_rid=to_rid, content=content)
        if print_sql:
            print(sql)
        r = self.execute(sql)[0]
        return r._OrientRecord__rid

    def get_create_rid(self, class_name: str, value_dict: dict, check_for=None, print_sql=False) -> str:
        """Return class_name.@rid by value_dict. Create record/insert if not exists."""
        rid = self.node_exists(
            class_name=class_name,
            value_dict=value_dict,
            check_for=check_for,
            print_sql=print_sql,
        )
        if not rid:
            rid = self.insert_record(class_name=class_name, value_dict=value_dict, print_sql=print_sql)
        return rid

    def update_correlative_edges(self) -> List[str]:
        """Create a reverse edge for every correlative edge."""
        updated_edges = []

        correlative_edges = self.query_class(class_name="correlative", with_rid=False, with_class=True)
        for c_edge in tqdm(correlative_edges, desc="Creating reverse correlative edges"):
            from_rid = c_edge.pop("in").get_hash()
            to_rid = c_edge.pop("out").get_hash()
            edge_class = c_edge.pop("class")
            c_edge["document"] = [doc.get_hash() for doc in c_edge["document"]]

            new_rid = self.create_edge(
                class_name=edge_class,
                from_rid=from_rid,
                to_rid=to_rid,
                value_dict=c_edge,
                if_not_exists=True,
            )
            updated_edges.append(new_rid)
        return updated_edges

    def update_document_info(self):
        """Update document metadata."""
        self.update_pmids()
        self.update_pmcids()

    def update_pmcids(self) -> int:
        """Add PMC ID to bel_relation if one exists."""
        update_sql_tmp = "UPDATE bel_relation SET pmc = '{}' WHERE pmid = {}"
        missing_pmc_sql = "SELECT distinct(pmid) as pmid from bel_relation WHERE pmc IS NULL"

        results = self.query(missing_pmc_sql)

        updated = 0
        if results is not None and "pmid" in results:
            pmids_with_missing_pmc = list(results["pmid"])

            id_dict = dict()
            for pmid_sublist in chunks(pmids_with_missing_pmc, size=200):
                pmid_string = ",".join([str(x) for x in pmid_sublist])
                url_filled = default_urls.NCBI_PMID.format(pmid_string)

                api_query_response = requests.get(url_filled)
                pmcids_json = json.loads(api_query_response.text)

                if "records" in pmcids_json.keys():
                    for record in pmcids_json["records"]:
                        if "pmid" in record.keys():
                            if "pmcid" in record.keys():
                                id_dict[record["pmid"]] = record["pmcid"]

                            else:
                                id_dict[record["pmid"]] = None

            for pmid, pmc in tqdm(id_dict.items(), desc="Updating PMC IDs"):
                if pmc:
                    update_sql_filled = update_sql_tmp.format(pmc, pmid)
                    r = self.execute(update_sql_filled)
                    updated += len(r)

        return updated

    def update_pmids(self, edge_name="bel_relation"):
        """Update PMID metadata for all edges of the specified edge_name."""
        sql_missing_citations = """Select distinct(pmid)
                                    as pmid from {}
                                    where
                                        pmid IS NOT NULL and
                                        pmid > 0 and
                                        citation.pubdate IS NULL"""
        r = self.execute(sql_missing_citations.format(edge_name))
        pmids = [x.oRecordData["pmid"] for x in r]

        updated = 0

        # length of PMID around 8-9 digits: 150*10=1500 + url with 103 => ~1600
        # max length of url = 2,083
        if pmids:
            chunk_size = 150
            total = len(pmids) // chunk_size + 1

            for pmid_chunk in tqdm(
                chunks(pmids, size=chunk_size),
                total=total,
                desc=f"Update PMID citations in {edge_name}",
            ):
                try:
                    updated += self._query_ncbi(pmid_chunk, edge_name)

                except KeyError as e:
                    logger.error("KeyError occurred during parsing. See logs for full description.")
                    logger.info(e)

        return updated

    def _query_ncbi(self, pmid_chunk: GeneratorType, edge_name: str):
        """Query NCBI for publication metadata."""
        sql_template = "Update {} set citation = {} where pmid = {}"

        sql_update_mesh_terms = "Update {} set annotation.mesh = {} where pmid = {}"
        sql_update_mesh_substances = "Update {} set annotation.substances = {} where pmid = {}"

        nameset = {"LastName", "Initials"}

        start_time = time.time()
        url = default_urls.NCBI_MESH + ",".join([str(x) for x in pmid_chunk])
        xml_pubmed = requests.get(url)
        parsed_data = xmltodict.parse(xml_pubmed.text)

        updated = 0
        if "PubmedArticleSet" in parsed_data:
            medline_citations = parsed_data["PubmedArticleSet"]["PubmedArticle"]

            if isinstance(medline_citations, OrderedDict):
                medline_citations = [
                    medline_citations,
                ]

            for medlineCitation in medline_citations:
                mc = medlineCitation["MedlineCitation"]
                data = {"type": "PubMed", "ref": mc["PMID"]["#text"]}
                article = mc["Article"]

                # Authors
                author_list = []
                if "AuthorList" in article:  # Authors not always listed
                    authors = article["AuthorList"]["Author"]

                    if isinstance(authors, OrderedDict):
                        authors = [authors]

                    for author in authors:
                        if isinstance(author, OrderedDict) and nameset.issubset(author.keys()):
                            author_list.append(author["LastName"] + " " + author["Initials"])

                data["author_list"] = author_list

                if author_list:
                    data["last_author"] = author_list[-1]

                # Journal
                data["full_journal_name"] = mc["Article"]["Journal"]["Title"]

                # Article Date
                article_date = mc.get("DateCompleted", mc.get("DateRevised"))
                if article_date:
                    ad = article_date.values()
                    data["pub_date"] = "-".join(ad)
                    data["pub_year"] = int(list(ad)[0])

                # Title
                data["title"] = mc["Article"]["ArticleTitle"]

                # DOI
                if "ELocationID" in mc["Article"]:
                    eids = mc["Article"]["ELocationID"]

                    if isinstance(eids, OrderedDict):
                        eids = [mc["Article"]["ELocationID"]]

                    for eid in eids:
                        if eid.get("@EIdType") == "doi":
                            data["doi"] = eid["#text"]

                # MeSH Headings
                meshs = []
                if "MeshHeadingList" in mc:
                    for mesh in mc["MeshHeadingList"]["MeshHeading"]:
                        m = mesh["DescriptorName"]
                        meshs.append(m["#text"])

                # Associated chemicals
                substances = []
                if "ChemicalList" in mc:
                    chemicals = mc["ChemicalList"]["Chemical"]
                    if isinstance(chemicals, OrderedDict):
                        chemicals = [chemicals]
                    for chemical in chemicals:
                        substances.append(chemical["NameOfSubstance"]["#text"])

                data_json = json.dumps(data)
                sql = sql_template.format(edge_name, data_json, data["ref"])
                self.execute(sql)

                if meshs:
                    content_mesh = json.dumps(meshs)
                    sql_m = sql_update_mesh_terms.format(edge_name, content_mesh, data["ref"])
                    self.execute(sql_m)

                if substances:
                    content_substances = json.dumps(substances)
                    sql_s = sql_update_mesh_substances.format(edge_name, content_substances, data["ref"])
                    self.execute(sql_s)

                updated += 1
        # don't make NCBI angry
        if (time.time() - start_time) < 1:
            time.sleep(1)

        return updated

    @staticmethod
    def _standardize_column_names(columns: Iterable[str]) -> List[str]:
        """Standardize column names.

        Parameters
        ----------
        columns: Iterable[str]
            Iterable of columns names.
        """
        return [get_standard_name(x) for x in columns]

    def _standardize_dataframe(
        self,
        dataframe: pd.DataFrame,
        replace_nulls_with_nones: bool = True,
        standardize_column_names: bool = True,
        replace_minus_with_nones: bool = True,
    ) -> pd.DataFrame:
        if standardize_column_names:
            dataframe.columns = self._standardize_column_names(dataframe.columns)

        if replace_nulls_with_nones:
            dataframe.replace({np.nan: None}, inplace=True)

        if replace_minus_with_nones and any(dataframe.dtypes == "object"):
            dataframe.replace({"-": None}, inplace=True)

        return dataframe

    def import_dataframe(
        self,
        dataframe: pd.DataFrame,
        class_name: str,
        replace_nulls_with_nones: bool = True,
        standardize_column_names: bool = True,
        replace: bool = True,
    ) -> int:
        """Import dataframe into OrientDb class with name."""
        sql_temp = f"insert into `{class_name}` content {{}}"
        inserted = 0

        if replace:
            self.clear_class(class_name)

        dataframe = self._standardize_dataframe(dataframe, replace_nulls_with_nones, standardize_column_names)

        for row in tqdm(
            dataframe.to_dict(orient="records"),
            desc=f"Insert {class_name.upper()} data",
        ):
            sql = sql_temp.format(json.dumps(row))

            try:
                self.execute(sql)

            except PyOrientCommandException as pyorientdb_command_exception:
                logging.error("OrientDB SQL error:", sql, pyorientdb_command_exception)

            inserted += 1

        return inserted

    def batch_insert(
        self,
        dataframe: pd.DataFrame,
        database: str,
        chunk_size: int = 100,
        desc: str = None,
        standardize_column_names: bool = False,
        replace: bool = True,
        replace_nulls_with_nones: bool = False,
    ) -> int:
        """Adds rows of a dataframe into specified generic table in batches.

        Parameters
        ----------
        dataframe: pandas DataFrame
            A dataframe of information to be inserted into the generic table.
        database: str
            Name of the generic table to be inserted into.
        chunk_size: int (optional)
            Number of chunks to break the dataframe into for batching.
        desc: str
            A description for tqdm about what is being iterated through.
        standardize_column_names: bool
            If True (default=False), standardize column names.
        replace: bool
            If True (default), content of dataframe replaces old data.
        replace_nulls_with_nones: bool
            If True (default=False), replace numpy.nan with None (==null in OrientDB).
        Returns
        -------
        int
            Number of rows inserted into the table.
        """
        sql_temp = "INSERT INTO {} CONTENT {}"

        inserted = 0
        total = len(dataframe.index) // chunk_size + 1

        if standardize_column_names:
            dataframe.columns = self._standardize_column_names(dataframe.columns)

        if replace:
            self.clear_class(database)

        if replace_nulls_with_nones:
            dataframe.replace({pd.np.nan: None}, inplace=True)

        for chunk in tqdm(chunks(dataframe.index, chunk_size), total=total, desc=desc):
            batch_cmds = ["begin"]
            for i in chunk:
                sql = sql_temp.format(database, dataframe.loc[i].to_json())
                batch_cmds.append(sql)
            batch_cmds.append("commit retry 100")
            cmd = ";".join(batch_cmds)
            self.client.batch(cmd)
            inserted += len(chunk)

        return inserted

    def get_set_gene_rids_by_position(
        self,
        chromosome: str,
        position: int,
        gene_types=["mapped", "downstream", "upstream"],
    ) -> Dict[str, List[str]]:
        """Return dictionary of mapped gene by chromosal position.

        ALERT: creates new BEL HGNC gene is not exists.
        """
        gene_rids = defaultdict(list)
        sqls = dict()

        sqls[
            "mapped"
        ] = f"""Select symbol
                            from ensembl
                            where
                                start < {position} and
                                stop > {position} and
                                chromosome='{chromosome}' group by symbol"""

        sqls[
            "downstream"
        ] = f"""Select symbol
                            from ensembl
                            where
                                start > {position} and
                                chromosome='{chromosome}'
                            order by start limit 1"""

        sqls[
            "upstream"
        ] = f"""Select symbol
                            from ensembl
                            where
                                stop < {position} and
                                chromosome='{chromosome}'
                            order by stop desc limit 1"""

        for gene_type, sql in sqls.items():
            if gene_type in gene_types:
                results = self.engine.execute(sql)
                for (symbol,) in results.fetchall():
                    bel = f'g(HGNC:"{symbol}")'
                    data = {
                        "name": symbol,
                        "namespace": "HGNC",
                        "bel": bel,
                        "pure": True,
                    }
                    gene_rid = self.get_create_rid("gene", value_dict=data, check_for="bel")
                    gene_rids[gene_type] += [gene_rid]

        return gene_rids

    def class_is_descendant_of(self, child_name: str, descendant_name: str) -> bool:
        """Returns True if child_name is a child class of descendent_name."""
        sql = "Select superClasses from (select expand(classes) from metadata:schema) where name='{}'"
        parents_exists = self.execute(sql.format(child_name))
        if parents_exists:
            data = parents_exists[0].oRecordData
            if data and "superClasses" in data:
                parent_names = data["superClasses"]

                if descendant_name in parent_names:
                    return True
                else:
                    for parent_name in parent_names:
                        return self.class_is_descendant_of(parent_name, descendant_name)
        return False

    def class_has_children(self, class_name) -> bool:
        """Checks if class_name has children that inherit from class_name."""
        sql = f"""Select count(name) from
                    (select expand(classes) from metadata:schema)
                  where '{class_name}' in superClasses"""
        return self.execute(sql)[0].oRecordData["count"] > 0

    def get_child_classes(self, class_name) -> List[str]:
        """Get list of child classes for given class_name."""
        sql = f"Select name from (select expand(classes) from metadata:schema) where '{class_name}' in superClasses"
        return [x.oRecordData["name"] for x in self.execute(sql)]

    def get_leaf_classes_of(self, class_name: str) -> List[str]:
        """Return list of children classes for the given class_name."""
        childs = self.get_child_classes(class_name)

        leafs = []
        for child in childs:
            if not self.class_has_children(child):
                leafs.append(child)
            else:
                leafs += self.get_leaf_classes_of(child)
        return leafs

    def insert(self) -> Dict[str, int]:
        """Check if files missing for download or generic table empty. If True then insert data."""
        inserted = {}
        downloaded = self.download()
        if any(downloaded.values()) or not all(self.number_of_generics.values()):
            inserted_entries = self.clear_and_import_data()
            inserted.update(inserted_entries)
        return inserted

    def update(self) -> None:
        """Check generics and update BEL interactions."""
        self.insert()
        self.update_bel()

    def update_bel(self) -> None:
        """Delete and update all class specific edges."""
        logger.info(f"Clear edges for {self.biodb_name}")
        self.clear_edges()
        logger.info(f"Update interactions for {self.biodb_name}")

        try:
            self.update_interactions()
            logger.info(f"Successfully updated interactions for {self.biodb_name.upper()}")

        except Exception as e:
            logging.error(e, exc_info=e)

    def delete_nodes_with_no_edges(self, class_name=None) -> int:
        """Delete all nodes without any edges."""
        if isinstance(class_name, str) and not self.class_exists(class_name):
            wtext = (
                f"You try to delete nodes with no edges from class {class_name}, "
                f"but node class {class_name} not exits."
            )
            logger.warning(wtext)
            return 0
        else:
            class_name = class_name if class_name is not None else "V"
            return self.execute(f"Delete VERTEX {class_name} where both().size() = 0")[0]

    def get_pure_symbol_rids_dict_in_bel_context(self, class_name="protein", namespace="HGNC") -> Dict[str, str]:
        """Return dictionary with HGNC names as key and OrientDB @rid as value.

        Applies to all pure nodes in graph with class name directly or indirectly involved in BEL stmt.
        This method could be helpful to avoid graph explosion.
        """
        # only include symbols which are also part of a BEL statement to avoid explosion of graph
        # Following sql also includes any modification of proteins

        # all pure nodes with bel_relation
        sql = (
            "Select name as symbol, @rid.asString() as rid from "
            + class_name
            + " where pure=true and namespace='"
            + namespace
            + "' and name in (Select name from (match { class:"
            + class_name
            + ", as: p, where:(namespace='"
            + namespace
            + "') }.(bothE('bel_relation'){ where: (document IS NOT NULL) }) return distinct(p.name) as name))"
        )
        return {r["symbol"]: r["rid"] for r in self.query_get_dict(sql)}

    def get_pure_uniprots_in_bel_context(self) -> Set[str]:
        """Returns a list of all uniprot accessions in BEL annotation context."""
        sql = (
            "match {class:protein, where:(uniprot IS NOT NULL), as:p}.(bothE('bel_relation')"
            "{class: bel_relation,where:(document IS NOT NULL)}) return distinct(p.uniprot) as uniprot"
        )
        return {x["uniprot"] for x in self.query_get_dict(sql)}

    def get_pure_symbol_rid_df_in_bel_context(self, class_name="protein", namespace="HGNC") -> pd.DataFrame:
        """Return dictionary with gene symbols as keys and node rIDs as values."""
        r = self.get_pure_symbol_rids_dict_in_bel_context(class_name=class_name, namespace=namespace)
        return pd.DataFrame(r.items(), columns=["symbol", "rid"])

    def get_pure_symbol_rids_dict(self, class_name="protein", namespace="HGNC") -> Dict[str, str]:
        """Return dictionary with protein name as keys and node rIDs as values."""
        results = self.query_class(class_name, pure=True, namespace=namespace)
        return {r["name"]: r["rid"] for r in results}

    def get_pure_rid_by_uniprot(self, uniprot: str):
        """Get rIDs of node based on UniProt ID."""
        sql = f"Select @rid.asString() as rid from protein where pure = true and uniprot='{uniprot}' limit 1"
        results = self.query_get_dict(sql)
        if results:
            return results[0]["rid"]

    def get_pure_uniprot_rid_dict_in_bel_context(self) -> Dict[str, str]:
        """Return dictionary with UniProt accession id as key and OrientDB @rid as value.

        Applies to all pure nodes in graph with class name directly or indirectly involved in BEL statement.
        This method could be helpful to avoid graph explosion.
        """
        # only include proteins which are also part of a BEL statement to avoid explosion of graph

        sql = """Select uniprot, @rid.asString() as rid from protein where pure=true and uniprot in (
        Select unionall(uniprot_list).asSet() as all_uniprots from (select unionall(in.uniprot, out.uniprot).asSet() as
        uniprot_list from bel_relation where document IS NOT NULL
        and (in.uniprot IS NOT NULL or out.uniprot IS NOT NULL)))"""

        return {r["uniprot"]: r["rid"] for r in self.query_get_dict(sql)}

    def get_pure_uniprot_rids_dict(self):
        """Return dictionary with UniProt IDs as keys and node rIDs as values."""
        sql = "Select uniprot, @rid.asString() as rid from protein where uniprot IS NOT NULL and pure=true"
        results = self.query_get_dict(sql)
        return {r["uniprot"]: r["rid"] for r in results}



"""ClinicalTrails.gov API methods."""
import json

from flask import request

from ebel.manager.rdbms.models import clinical_trials_gov as ct
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_paginated_query_result,
                                  _get_terms_from_model_starts_with)


def get_ct_by_nct_id():
    """Get CT entry based on given NCT ID."""
    nct_id = request.args.get("nct_id")
    c = RDBMS.get_session().query(ct.ClinicalTrialGov).filter_by(nct_id=nct_id).first()
    if c:
        return c.as_dict()


# not used, uses too much memory
# def get_ct_by_contains_in_summary_or_description_or_title():
#     search_term = json.loads(request.data)['search_term']
#     search_str = f"%{search_term}%"
#     results = session.query(ct).filter(or_(
#         ct.ClinicalTrialGov.brief_title.like(search_str),
#         ct.ClinicalTrialGov.official_title.like(search_str),
#         ct.ClinicalTrialGov.brief_summary.like(search_str),
#         ct.ClinicalTrialGov.detailed_description.like(search_str)
#     ))
#     return [x.as_dict() for x in results]


def get_ct_by_mesh_term():
    """Get CT entry based on given mesh term."""
    mesh_term = request.args.get("mesh_term", None)
    if not mesh_term:
        return {"error": "Mesh term is required."}
    query = RDBMS.get_session().query(ct.MeshTerm).filter(ct.MeshTerm.mesh_term.like(mesh_term))

    return _get_paginated_query_result(query)


def get_ct_by_intervention():
    """Get CT entry based on given intervention."""
    intervention_name = request.args.get("intervention_name", None)
    intervention_type = request.args.get("intervention_type", None)
    if not (intervention_name or intervention_type):
        return {"error": "At least name or type is required."}
    query = RDBMS.get_session().query(ct.Intervention)

    if intervention_type:
        query = query.filter_by(intervention_type=intervention_type)
    if intervention_type:
        query = query.filter(ct.Intervention.intervention_name.like(intervention_name))

    return _get_paginated_query_result(query)


def get_ct_by_keyword():
    """Get CT entry based on given keyword."""
    keyword = request.args.get("keyword", None)
    if not keyword:
        return {"error": "keyword is required."}
    query = RDBMS.get_session().query(ct.Keyword).filter(ct.Keyword.keyword.like(keyword))

    return _get_paginated_query_result(query)


def get_ct_by_condition():
    """Get CT entry based on given condition."""
    condition = request.args.get("condition", None)
    if not condition:
        return {"error": "Condition is required."}
    query = RDBMS.get_session().query(ct.Condition).filter(ct.Condition.condition.like(condition))

    return _get_paginated_query_result(query)


def get_mesh_term_starts_with():
    """Get CT entry based on beginning of mesh term string."""
    return _get_terms_from_model_starts_with(form_field="mesh_term", sa_column=ct.MeshTerm.mesh_term)


def get_keyword_starts_with():
    """Get CT entry based on beginning of keyword string."""
    return _get_terms_from_model_starts_with(form_field="keyword", sa_column=ct.Keyword.keyword)


def get_condition_starts_with():
    """Get CT entry based on beginning of condition string."""
    return _get_terms_from_model_starts_with(form_field="condition", sa_column=ct.Condition.condition)



"""EnsEMBL API methods."""
from ebel.manager.rdbms.models import ensembl
from ebel.web.api.ebel.v1 import _get_data


def get_ensembl():
    """Get generic ensembl entry."""
    return _get_data(ensembl.Ensembl)



"""Reactome RDBMS model definition."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Reactome(Base):
    """Class definition for the reactome table."""

    __tablename__ = "reactome"
    id = Column(Integer, primary_key=True)
    identifier = Column(String(50), index=True)
    uniprot_accession = Column(String(50), index=True)
    organism = Column(String(255))
    name = Column(String(255))
    evidence_type = Column(String(255))

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""EnsEMBL API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestEnsembl:
    def test_get_ensembl(self, client):
        response = client.get(
            "api/v1/ebel/ensembl/ensembl?enst=ENST00000436584&chromosome=19", content_type="application/json"
        )
        output = format_response_data(response)
        expected_results = {
            "chromosome": "19",
            "enst": "ENST00000436584",
            "gene_id": "ENSG00000105383.15",
            "gene_id_short": "ENSG00000105383",
            "hgnc_id": "HGNC:1659",
            "id": 29665,
            "orientation": 1,
            "start": 51225064,
            "stop": 51236572,
            "symbol": "CD33",
            "version": 38,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Unit tests for the enrich methods."""

from ..constants import TEST_JSON, test_client

bel = test_client


class TestEnrich:
    # TODO: Erasing and reimporting causes an error when these tests are run with test_bel.py
    # Make sure test DB is populated
    bel.clear_all_nodes_and_edges()
    bel.clear_documents()

    assert all([count == 0 for count in bel.number_of_nodes.values()])
    assert all([count == 0 for count in bel.number_of_edges.values()])

    bel.import_json(
        input_path=TEST_JSON,
        update_from_protein2gene=True,
        extend_graph=False,
    )

    def test_uniprot_enrich(self):
        updated = bel.uniprot.update_bel()
        assert isinstance(updated, dict)
        assert all([species in updated for species in ("HGNC", "MGI", "RGD")])
        assert all([isinstance(val, int) for val in updated.values()])
        assert updated["HGNC"] == 3

    def test_hgnc_enrich(self):
        updated = bel.hgnc.update_bel()
        assert isinstance(updated, int)
        assert updated == 4

    def test_chebi_enrich(self):
        updated = bel.chebi.update_bel()
        assert isinstance(updated, int)
        assert updated == 1

    def test_gwas_enrich(self):
        updated = bel.gwas_catalog.update_bel()
        assert isinstance(updated, dict)
        assert updated == {"has_mapped_snp_gc": 1141, "has_downstream_snp_gc": 401, "has_upstream_snp_gc": 394}

    def test_clinvar_enrich(self):
        updated = bel.clinvar.update_bel()
        assert isinstance(updated, int)
        assert updated == 2326

    def test_reactome_enrich(self):
        updated = bel.reactome.update_interactions()
        assert isinstance(updated, int)
        assert updated == 3

    def test_stringdb_enrich(self):
        updated = bel.stringdb.update_interactions()
        assert isinstance(updated, dict)
        assert updated == {"interactions": 43, "actions": 152}

    def test_biogrid_enrich(self):
        updated = bel.biogrid.update_interactions()
        assert isinstance(updated, int)
        assert updated == 0

    def test_mirtarbase_enrich(self):
        updated = bel.mirtarbase.update_interactions()
        assert isinstance(updated, int)
        assert updated == 0

    def test_pc_enrich(self):
        updated = bel.pathway_commons.update_interactions()
        assert isinstance(updated, dict)
        assert updated == {
            "controls-transport-of": 19,
            "controls-expression-of": 71,
            "controls-phosphorylation-of": 12,
        }

    def test_disgenet_enrich(self):
        updated = bel.disgenet.update_interactions()
        assert isinstance(updated, int)
        assert updated == 2948

    def test_kegg_enrich(self):
        updated = bel.kegg.update_interactions()
        assert isinstance(updated, int)
        assert updated == 0

    def test_drugbank_enrich(self):
        updated = bel.drugbank.update_interactions()
        assert isinstance(updated, int)
        assert updated == 3

    def test_iuphar_enrich(self):
        updated = bel.iuphar.update_interactions()
        assert isinstance(updated, int)
        assert updated == 3

    def test_nsides_enrich(self):
        updated = bel.nsides.update_bel()
        assert isinstance(updated, int)
        assert updated == 297

    def test_ct_enrich(self):
        # currently disabled
        pass



"""IUPHAR API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS

example_result = {
    "action": "Inhibition",
    "action_comment": None,
    "affinity_high": None,
    "affinity_low": None,
    "affinity_median": 8.77,
    "affinity_units": "pIC50",
    "approved_drug": "0",
    "assay_description": None,
    "concentration_range": None,
    "endogenous": False,
    "id": 9631,
    "ligand": "ACY-738",
    "ligand_context": None,
    "ligand_gene_symbol": None,
    "ligand_id": 11120,
    "ligand_pubchem_sid": 434122180,
    "ligand_species": None,
    "original_affinity_high_nm": None,
    "original_affinity_low_nm": None,
    "original_affinity_median_nm": 1.7,
    "original_affinity_relation": "=",
    "original_affinity_units": "IC50",
    "primary_target": False,
    "pubmed_id": "23954848",
    "receptor_site": None,
    "selectivity": "Not Determined",
    "target": "histone deacetylase 6",
    "target_ensembl_gene_id": "ENSG00000094631",
    "target_gene_symbol": "HDAC6",
    "target_id": 2618,
    "target_ligand": None,
    "target_ligand_ensembl_gene_id": None,
    "target_ligand_gene_symbol": None,
    "target_ligand_id": None,
    "target_ligand_pubchem_sid": None,
    "target_ligand_uniprot": None,
    "target_species": "Human",
    "target_uniprot": "Q9UBN7",
    "type": "Inhibitor",
}


class TestIuphar:
    def test_get_interaction(self, client):
        response = client.get(
            "api/v1/iuphar/interaction?target=CD33&target_id=2601&target_gene_symbol=CD33&target_uniprot=P20138&target_ensembl_gene_id=ENSG00000105383&target_species=Human&ligand=gemtuzumab%20ozogamicin&ligand_id=6775&ligand_pubchem_sid=178103381&approved_drug=1&type=Antibody&action=Binding&selectivity=Selective&primary_target=true&original_affinity_relation=%3D&pubmed_id=10720144&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "action": "Binding",
            "action_comment": None,
            "affinity_high": None,
            "affinity_low": None,
            "affinity_median": None,
            "affinity_units": "-",
            "approved_drug": "1",
            "assay_description": None,
            "concentration_range": None,
            "endogenous": False,
            "id": 4512,
            "ligand": "gemtuzumab ozogamicin",
            "ligand_context": None,
            "ligand_gene_symbol": None,
            "ligand_id": 6775,
            "ligand_pubchem_sid": 178103381,
            "ligand_species": None,
            "original_affinity_high_nm": None,
            "original_affinity_low_nm": None,
            "original_affinity_median_nm": None,
            "original_affinity_relation": "=",
            "original_affinity_units": "-",
            "primary_target": True,
            "pubmed_id": "10720144",
            "receptor_site": None,
            "selectivity": "Selective",
            "target": "CD33",
            "target_ensembl_gene_id": "ENSG00000105383",
            "target_gene_symbol": "CD33",
            "target_id": 2601,
            "target_ligand": None,
            "target_ligand_ensembl_gene_id": None,
            "target_ligand_gene_symbol": None,
            "target_ligand_id": None,
            "target_ligand_pubchem_sid": None,
            "target_ligand_uniprot": None,
            "target_species": "Human",
            "target_uniprot": "P20138",
            "type": "Antibody",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_ligandby_by_id(self, client):
        response = client.get("api/v1/iuphar/ligand/by_id?id=1", content_type="application/json")
        output = format_response_data(response)

        expected_result = {
            "approved": None,
            "gto_immu_pdb": None,
            "gto_mpdb": None,
            "id": 1,
            "inchi": "InChI=1S/C22H26FN3O4/c23-17-6-4-16(5-7-17)22(28)24-8-9-25-10-12-26(13-11-25)19-2-1-3-20-21(19)29-15-18(14-27)30-20/h1-7,18,27H,8-15H2,(H,24,28)/t18-/m0/s1",
            "inchi_key": "NYSDRDDQELAVKP-SFHVURJKSA-N",
            "inn": "flesinoxan",
            "iupac_name": "4-fluoro-N-[2-[4-[(3S)-3-(hydroxymethyl)-2,3-dihydro-1,4-benzodioxin-8-yl]piperazin-1-yl]ethyl]benzamide",
            "labelled": None,
            "name": "flesinoxan",
            "pubchem_cid": "57347",
            "pubchem_sid": 135650267,
            "radioactive": None,
            "smiles": "OC[C@H]1COc2c(O1)cccc2N1CCN(CC1)CCNC(=O)c1ccc(cc1)F",
            "species": None,
            "synonyms": "(+)-flesinoxan|DU-29,373",
            "type": "Synthetic organic",
            "uniprot_id": None,
            "withdrawn": None,
        }
        assert isinstance(output, dict)
        assert output == expected_result

    def test_get_interaction_by_target_uniprot(self, client):
        response = client.get(
            "api/v1/iuphar/interaction/by_target_uniprot?target_uniprot=P20138&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)

        expected_cols = example_result.keys()
        assert output[NUM_RESULTS] >= 3
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])

    def test_get_interaction_by_target_gene_symbol(self, client):
        response = client.get(
            "api/v1/iuphar/interaction/by_target_gene_symbol?target_gene_symbol=HDAC6&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)

        expected_cols = example_result.keys()
        assert output[NUM_RESULTS] >= 3
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])



"""General methods used by e(BE:L) modules."""
import configparser
import gzip
import hashlib
import os.path
import re
import shutil
from types import GeneratorType
from typing import Iterable, List, Union

from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm import sessionmaker

from ebel import defaults
from ebel.config import get_config_value, write_to_config
from ebel.constants import DATA_DIR
from ebel.defaults import CONN_STR_DEFAULT


class BelRdb(object):
    """Singleton for Relational Database Connection."""

    __instance = None

    def __new__(cls):
        """Define new object."""
        if BelRdb.__instance is None:
            BelRdb.__instance = object.__new__(cls)
            connection_string = _get_connection_string()
            dialect = re.search(r"^(\w+)(\+\w+)?:", connection_string).group(1)
            if dialect == "mysql":
                utf8mb4 = "charset=utf8mb4"
                if utf8mb4 not in connection_string:
                    connection_string = connection_string + f"?{utf8mb4}"
                BelRdb.__instance.engine = create_engine(connection_string, pool_size=30, max_overflow=10)
            else:
                BelRdb.__instance.engine = create_engine(connection_string)
            BelRdb.__instance.session = sessionmaker(bind=BelRdb.__instance.engine)()
        return BelRdb.__instance


def _get_connection_string():
    """Get the sqlalchemy connection string from config file, sets the default string if not there."""
    return get_config_value("DATABASE", "sqlalchemy_connection_string", CONN_STR_DEFAULT)


def _get_engine() -> Engine:
    """Get SQLAlchmey engine."""
    connection_string = _get_connection_string()
    engine = create_engine(connection_string, pool_size=30, max_overflow=10)
    return engine


def _get_session_object():
    return sessionmaker(bind=_get_engine())


def get_session():
    """Return session object."""
    return _get_session_object()()


def chunks(parsable: Iterable, size: int = 100) -> GeneratorType:
    """Generate a list of chunks with specific size [default=100]."""
    chunk = []
    counter = 0
    for e in parsable:
        counter += 1
        chunk.append(e)
        if counter == size:
            yield chunk
            chunk = []
            counter = 0
    yield chunk


def get_standard_name(name: str) -> str:
    """Return standard name."""
    part_of_name = [x for x in re.findall("[A-Z]*[a-z0-9]*", name) if x]
    new_name = "_".join(part_of_name).lower()
    if re.search(r"^\d+", new_name):
        new_name = "_" + new_name
    return new_name


def md5(file_path):
    """Calculate md5 for file."""
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_file_path(url: str, biodb: str):
    """Get standard file path by file_name and DATADIR."""
    file_name = os.path.basename(url)
    bio_db_dir = os.path.join(DATA_DIR, biodb)
    os.makedirs(bio_db_dir, exist_ok=True)
    return os.path.join(bio_db_dir, file_name)


def get_file_name(url_or_path):
    """Get standard file path by file_name and DATADIR."""
    return os.path.basename(url_or_path)


def gunzip(file_path: str, file_path_gunzipped: str):
    """Gunzip a file."""
    with gzip.open(file_path, "rb") as f_in:
        with open(file_path_gunzipped, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


def get_disease_trait_keywords_from_config(traits: Union[str, list] = None, overwrite: bool = False) -> List[str]:
    """Interface with the e(BE:L) config file.

    This reads and returns keywords found in the 'SNP_RELATED_TRAITS' section or it will create the section using
    the passed traits.

    Parameters
    ----------
    traits: Union[str, List[str]]
        The list of pathologies/pathogens.
    overwrite: [bool, optional]
        Can choose to overwrite the default disease_trait_keyword value in the config file.

    Returns
    -------
    If the config file has the 'SNP_RELATED_TRAITS' section, the method returns the value(s) stored in the
    "keyword" section. If no section or config file exists, returns the default value: ['Alzheimer'].
    """
    section_name = "SNP_RELATED_TRAITS"
    option = "keyword"

    if traits and isinstance(traits, list):
        traits = ",".join([trait.strip() for trait in traits])

    if traits and isinstance(traits, str):
        traits = ",".join([trait.strip() for trait in traits.split(",")])  # Split, strip, recombine

    if os.path.isfile(defaults.config_file_path):  # If config file exists
        cfg = configparser.ConfigParser()
        cfg.read(defaults.config_file_path)

        if cfg.has_section(section_name) and not overwrite:  # If there is a section for the default value
            traits = cfg.get(section_name, option)  # Get keyword from config

        elif traits and overwrite:  # Traits passed and overwrite enabled
            write_to_config(section_name, option, traits)

        elif not cfg.has_section(section_name) and traits:  # No section but traits passed
            write_to_config(section_name, option, traits)

    else:  # If no config file then write a new one with the section/option/value
        write_to_config(section_name, option, traits)

    if not isinstance(traits, list):
        traits = traits.split(",")

    return traits if traits is not None else []



"""API queries for collecting contraidctory edges."""

from math import ceil
from typing import Dict

from flask import request

from ebel import Bel
from ebel.web.api.ebel.v1 import _get_pagination


def _get_where_sql(where_dict: Dict[str, str]) -> str:
    where = ""
    if any(where_dict.values()):
        where = ", WHERE:(" + " AND ".join([f"{k} = '{v}'" for k, v in where_dict.items() if v]) + ")"
    return where


def get_contradictory_edges():
    """Return list of nodes with a given namespace."""
    b = Bel()

    subject_class = request.args.get("subject_class", "bel")
    object_class = request.args.get("object_class", "bel")

    sub, obj = {}, {}
    sub["namespace"] = request.args.get("subject_namespace")
    sub["name"] = request.args.get("subject_name")
    obj["namespace"] = request.args.get("object_namespace")
    obj["name"] = request.args.get("object_name")

    pagination = _get_pagination()
    match = (
        f"match {{class:{subject_class}, as:subject_1 {_get_where_sql(sub)} }}"
        ".outE('directly_increases', 'increases'){as:relation_1}"
        f".inV(){{class:{object_class}, as:object {_get_where_sql(obj)} }}"
        ".inE('decreases', 'directly_decreases'){as:relation_2}"
        f".outV(){{class:{subject_class}, as:subject_2, where:($matched.subject_1=$currentMatch)}}"
        """ return
                subject_1.@rid.asString() as subject_rid,
                subject_1.@class.asString() as subject_class,
                subject_1.namespace as subject_namespace,
                subject_1.name as subject_name,
                subject_1.bel as subject_bel,
                relation_1.@rid.asString() as relation_1_rid,
                relation_1.@class as relation_1_class,
                relation_1.evidence as relation_1_evidence,
                relation_1.pmid as relation_1_pmid,
                relation_1.annotation as relation_1_annotation,
                object.@rid.asString() as object_rid,
                object.@class.asString() as object_class,
                object.namespace as object_namespace,
                object.name as object_name,
                object.bel as object_bel,
                relation_2.@rid.asString() as relation_2_rid,
                relation_2.@class as relation_2_class,
                relation_2.evidence as relation_2_evidence,
                relation_2.pmid as relation_2_pmid,
                relation_2.annotation as relation_2_annotation"""
    )

    number_of_results = b.execute(f"Select count(*) as number from ({match})")[0].oRecordData["number"]

    sql_pagination = f"Select * from ({match}) limit {pagination.page_size} skip {pagination.skip}"
    results = [x.oRecordData for x in b.execute(sql_pagination)]
    pages = ceil(number_of_results / pagination.page_size)

    return {
        "page": pagination.page,
        "page_size": pagination.page_size,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": results,
    }



"""Generic methods for the models submodule."""
from sqlalchemy import inspect


def object_as_dict(obj, exclude: list = []) -> dict:
    """Return object values as a dictionary."""
    return {c.key: getattr(obj, c.key) for c in inspect(obj).mapper.column_attrs if c.key not in exclude}



"""Unit tests for checking the URLs."""

import ftplib

import requests

from ebel.manager.orientdb.constants import (BIOGRID, CHEBI, CLINICAL_TRIALS,
                                             CLINVAR, DISGENET, DRUGBANK,
                                             ENSEMBL, GWAS_CATALOG, HGNC,
                                             INTACT, IUPHAR, KEGG, MIRTARBASE,
                                             NCBI, NSIDES, PATHWAY_COMMONS,
                                             PROTEIN_ATLAS, REACTOME, STRINGDB,
                                             UNIPROT)

from ..constants import test_client

bel = test_client


class TestUrls:
    """Goes through URLs for bioDBs and checks whether link is still valid."""

    @staticmethod
    def check_links(urls: dict):
        """Checks every link in dict is downloadable."""
        for url_name, url in urls.items():
            if url.startswith("ftp"):  # FTP conns need to be handled
                remove_prefix = url.split("://")[1]
                base_name, path = remove_prefix.split("/", 1)
                ftp_conn = ftplib.FTP(base_name)
                ftp_conn.login("anonymous", "anonymous")
                assert ftp_conn.size(path) > 0  # Check file is there
                ftp_conn.close()

            else:
                if "kegg" in url_name:  # Special case - returns plain text
                    resp = requests.get(url)
                    assert resp.ok
                    continue

                headers = requests.head(url, allow_redirects=True).headers
                if "Content-disposition" in headers:  # Check file attachment
                    assert "filename" in headers.get("Content-disposition")

                elif "Content-Encoding" in headers:
                    assert headers.get("Content-Encoding") == "gzip"  # Check if it's a gzip file

                else:
                    file_size = int(headers.get("Content-Length"))
                    assert file_size > 0

    def test_urls(self):
        """Iterates through BioDBs."""
        biodbs = {
            HGNC: bel.hgnc,
            CHEBI: bel.chebi,
            ENSEMBL: bel.ensembl,
            GWAS_CATALOG: bel.gwas_catalog,
            CLINVAR: bel.clinvar,
            UNIPROT: bel.uniprot,
            REACTOME: bel.reactome,
            STRINGDB: bel.stringdb,
            INTACT: bel.intact,
            BIOGRID: bel.biogrid,
            MIRTARBASE: bel.mirtarbase,
            PATHWAY_COMMONS: bel.pathway_commons,
            DISGENET: bel.disgenet,
            KEGG: bel.kegg,
            DRUGBANK: bel.drugbank,
            IUPHAR: bel.iuphar,
            NSIDES: bel.nsides,
            CLINICAL_TRIALS: bel.clinical_trials,
            PROTEIN_ATLAS: bel.protein_atlas,
            NCBI: bel.ncbi,
        }
        for db in biodbs.values():
            self.check_links(db.urls)



"""GWAS Catalog RDBMS model definition."""
from sqlalchemy import Column, Float, ForeignKey, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class GwasCatalog(Base):
    """Class definition for the gwascatalog table."""

    __tablename__ = "gwascatalog"
    id = Column(Integer, primary_key=True)
    date_added_to_catalog = Column(String(255))
    pubmedid = Column(Integer)
    first_author = Column(String(255))
    date = Column(String(255))
    journal = Column(String(255))
    link = Column(String(255))
    study = Column(Text)
    disease_trait = Column(String(255))
    initial_sample_size = Column(Text)
    replication_sample_size = Column(Text)
    region = Column(String(50))
    chr_id = Column(Text)
    chr_pos = Column(Text)
    reported_gene_s = Column(Text)
    mapped_gene = Column(Text)
    upstream_gene_id = Column(String(50))
    downstream_gene_id = Column(String(50))
    upstream_gene_distance = Column(Integer)
    downstream_gene_distance = Column(Integer)
    strongest_snp_risk_allele = Column(Text)
    snp = Column(Text)
    merged = Column(Integer)
    snp_id_current = Column(Text)
    context = Column(Text)
    intergenic = Column(Integer)
    risk_allele_frequency = Column(Text)
    p_value = Column(Float)
    pvalue_mlog = Column(Float)
    p_value_text = Column(Text)
    or_or_beta = Column(Float)
    _95_ci_text = Column(Text)
    platform_snps_passing_qc = Column(Text)
    cnv = Column(Text)

    snp_genes = relationship("SnpGene", back_populates="gwascatalog")

    def as_dict(self):
        """Convert object values to dictionary."""
        gwas_catalog = object_as_dict(self)
        gwas_catalog.update({"snp_genes": [x.ensembl_identifier for x in self.snp_genes]})
        return gwas_catalog


class SnpGene(Base):
    """Class definition for the gwascatalog_snpgene table."""

    __tablename__ = "gwascatalog_snpgene"
    id = Column(Integer, primary_key=True)
    ensembl_identifier = Column(String(100), nullable=False, index=True)
    gwascatalog_id = Column(Integer, ForeignKey("gwascatalog.id"))
    gwascatalog = relationship("GwasCatalog", back_populates="snp_genes")



"""OrientDB defaults."""
from enum import Enum
from typing import Dict


class OIndexType(Enum):
    """Allowed OrientDB Index types."""

    NOTUNIQUE = "notunique"
    UNIQUE = "unique"
    FULLTEXT = "fulltext"
    DICTIONARY = "dictionary"
    UNIQUE_HASH_INDEX = "unique_hash_index"
    NOTUNIQUE_HASH_INDEX = "notunique_hash_index"
    FULLTEXT_HASH_INDEX = "fulltext_hash_index"
    DICTIONARY_HASH_INDEX = "dictionary_hash_index"


class ODataType(Enum):
    """Allowed OrientDB Data types."""

    BOOLEAN = "BOOLEAN"
    BINARY = "BINARY"
    BYTE = "BYTE"
    DATE = "DATE"
    DATETIME = "DATETIME"
    DECIMAL = "DECIMAL"
    DOUBLE = "DOUBLE"
    EMBEDDED = "EMBEDDED"
    EMBEDDEDLIST = "EMBEDDEDLIST"
    EMBEDDEDSET = "EMBEDDEDSET"
    EMBEDDEDMAP = "EMBEDDEDMAP"
    FLOAT = "FLOAT"
    INTEGER = "INTEGER"
    LONG = "LONG"
    LINK = "LINK"
    LINKLIST = "LINKLIST"
    LINKSET = "LINKSET"
    LINKMAP = "LINKMAP"
    LINKBAG = "LINKBAG"
    STRING = "STRING"
    SHORT = "SHORT"


bel_func_short: Dict[str, str] = {
    "gmod": "gmod",
    "protein": "p",
    "abundance": "a",
    "micro_rna": "m",
    "rna": "r",
    "gene": "g",
    "activity": "act",
    "fragment": "frag",
    "pmod": "pmod",
    "location": "loc",
    "variant": "var",
    "complex": "complex",
    "reaction": "rxn",
    "reactants": "reactants",
    "products": "products",
    "pathology": "path",
    "degradation": "deg",
    "biological_process": "bp",
    "list": "list",
    "cell_secretion": "sec",
    "composite": "composite",
    "translocation": "tloc",
    "fusion_protein": "fus",
    "fusion_rna": "fus",
    "fusion_gene": "fus",
    "from_location": "fromLoc",
    "to_location": "toLoc",
    "cell_surface_expression": "surf",
    "population": "pop",
}

normalized_pmod: Dict[str, str] = {
    "ace": "Ac",
    "adr": "ADPRib",
    "add": '"ADP-rybosylation"',
    "far": "Farn",
    "ger": "Gerger",
    "gly": "Glyco",
    "hyd": "Hy",
    "isg": "ISG",
    "me0": "Me",
    "me1": "Me1",
    "mon": "monomethylation",
    "me2": "Me2",
    "me3": "Me3",
    "tri": "trimethylation",
    "myr": "Myr",
    "ned": "Nedd",
    "ngl": "NGlyco",
    "nit": "NO",
    "ogl": "OGlyco",
    "pal": "Palm",
    "pho": "Ph",
    "sul": "Sulf",
    "sup": "sulphation",
    "suh": "sulfonation",
    "sum": "sulphonation",
    "suy": "Sumo",
    "ubi": "Ub",
    "u48": "UbK48",
    "u63": "UbK63",
    "ubm": "UbMono",
    "ubp": "UbPoly",
    "pre": "Prenylation",  # added for BioGrid
    "dei": "de-ISGylation",  # added for BioGrid
    "fat": "FAT10ylation",
    "ufm": "Ufmylation",  # added for BioGrid
}

normalized_pmod_reverse = {v: k for k, v in normalized_pmod.items()}


class BelPmod:
    """Protein modification definitions."""

    ACE = normalized_pmod["ace"]
    ADR = normalized_pmod["adr"]
    ADD = normalized_pmod["add"]
    FAR = normalized_pmod["far"]
    GER = normalized_pmod["ger"]
    GLY = normalized_pmod["gly"]
    HYD = normalized_pmod["hyd"]
    ISG = normalized_pmod["isg"]
    ME0 = normalized_pmod["me0"]
    ME1 = normalized_pmod["me1"]
    MON = normalized_pmod["mon"]
    ME2 = normalized_pmod["me2"]
    ME3 = normalized_pmod["me3"]
    TRI = normalized_pmod["tri"]
    MYR = normalized_pmod["myr"]
    NED = normalized_pmod["ned"]
    NGL = normalized_pmod["ngl"]
    NIT = normalized_pmod["nit"]
    OGL = normalized_pmod["ogl"]
    PAL = normalized_pmod["pal"]
    PHO = normalized_pmod["pho"]
    SUL = normalized_pmod["sul"]
    SUP = normalized_pmod["sup"]
    SUH = normalized_pmod["suh"]
    SUM = normalized_pmod["sum"]
    SUY = normalized_pmod["suy"]
    UBI = normalized_pmod["ubi"]
    U48 = normalized_pmod["u48"]
    U63 = normalized_pmod["u63"]
    UBM = normalized_pmod["ubm"]
    UBP = normalized_pmod["ubp"]
    PRE = normalized_pmod["pre"]  # added for BioGrid
    DEI = normalized_pmod["dei"]  # added for BioGrid
    FAT = normalized_pmod["fat"]
    UFM = normalized_pmod["ufm"]  # added for BioGrid



"""Unit tests for BEL grammar validation."""

import os
import pathlib

import pandas as pd

from ebel.validate import validate_bel_file

VALIDATION_TEST_DIR = pathlib.Path(__file__).parent.absolute()
TEST_DATA_DIR = os.path.join(VALIDATION_TEST_DIR, "..", "data")

FUNC_REL_TEST_FILE = os.path.join(TEST_DATA_DIR, "function_relation_bel_tests.bel")
TERM_TEST_FILE = os.path.join(TEST_DATA_DIR, "terminology_bel_tests.bel")
BEL2_1_TEST_FILE = os.path.join(TEST_DATA_DIR, "bel_2_1_test_statements.bel")
CORRECT_BEL_FILE = os.path.join(TEST_DATA_DIR, "correct_statements.bel")


class TestValidation:
    """Class for checking the BEL grammar and syntax validation."""

    @staticmethod
    def check_files(grammar_file: str, version: str):
        """Wrapper method to perform the checks."""
        error_report = "function_errors.xlsx"

        with open(grammar_file, "r") as bel_file:
            content = bel_file.readlines()

        num_false_statements = 0
        false_lines = set()
        for line_number, line_text in enumerate(content):
            if line_text == "# Return False\n":
                num_false_statements += 1
                false_lines.add(line_number + 2)  # Add 2 - index starts at 0 and stmt is line after "Return False"

        validate_bel_file(grammar_file, bel_version=version, json_file=False, reports=error_report)

        assert os.path.isfile(error_report)
        error_df = pd.read_excel(error_report)
        num_found_errors = len(error_df)
        assert num_false_statements == num_found_errors
        reported_error_lines = set(error_df["line_number"])
        assert false_lines == reported_error_lines  # Should be the same due to matching line reports
        os.remove(error_report)
        assert not os.path.isfile(error_report)

    def test_function_relation_validation_2_0(self):
        """Check basic grammar and syntax rules using the BEL 2.0 grammar file."""
        self.check_files(grammar_file=FUNC_REL_TEST_FILE, version="2")

    def test_function_relation_validation_2_1(self):
        """Check basic grammar and syntax rules using the BEL 2.1 grammar file."""
        self.check_files(grammar_file=FUNC_REL_TEST_FILE, version="2_1")

    def test_2_1_additions(self):
        """Goes through the new additions to the grammar in BEL 2.1 and verifies they are acceptable."""
        self.check_files(grammar_file=BEL2_1_TEST_FILE, version="2_1")

    def test_json_generation(self):
        error_file_path = os.path.join(TEST_DATA_DIR, "function_errors.xlsx")
        assert not os.path.isfile(error_file_path)
        json_file_path = CORRECT_BEL_FILE + ".json"
        assert not os.path.isfile(json_file_path)

        validate_bel_file(CORRECT_BEL_FILE, bel_version="2_1", json_file=True, reports=error_file_path)
        assert not os.path.isfile(error_file_path)
        assert os.path.isfile(json_file_path)
        os.remove(json_file_path)
        assert not os.path.isfile(json_file_path)



"""NSIDES module."""

import logging
import os
import tarfile
from typing import Dict, Optional

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.constants import RID
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import OFFSIDES, ONSIDES
from ebel.manager.rdbms.models import nsides
from ebel.tools import get_file_path

logger = logging.getLogger(__name__)
pd.options.mode.chained_assignment = None


class Nsides(odb_meta.Graph):
    """Drug side effects and drug-drug interactions were mined from publicly available data.

    Offsides is a database
    of drug side effects that were found, but are not listed on the official FDA label. Twosides is the only
    comprehensive database drug-drug-effect relationships. Over 3,300 drugs and 63,000 combinations connected to
    millions of potential adverse reactions.

    Onsides is a database of adverse reactions and boxed warnings extracted from the FDA structured product labels.
    All labels available to download from DailyMed
    (https://dailymed.nlm.nih.gov/dailymed/spl-resources-all-drug-labels.cfm) as of April 2022 were processed in
    this analysis. In total 2.7 million adverse reactions were extracted from 42,000 labels for just
    under 2,000 drug ingredients or combination of ingredients.

    More information can be found on http://tatonettilab.org/offsides/
    """

    def __init__(self, client: OrientDB = None):
        """Init nSIDES."""
        self.client = client
        self.biodb_name = "nsides"
        self.urls = {OFFSIDES: urls.OFFSIDES, ONSIDES: urls.ONSIDES}
        super().__init__(
            edges=odb_structure.nsides_edges,
            nodes=odb_structure.nsides_nodes,
            tables_base=nsides.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class.

        Columns:

        drug_rxnorn_id 	integer 	RxNorm CUI (RxCUI) of each drug
        drug_concept_name 	string 	String name of each drug
        condition_meddra_id 	integer 	MedDRA code of each (adverse event) condition
        condition_concept_name 	string 	String name of each condition
        source  string  The database the information was derived from: OFFSIDES or OnSIDES
        A 	integer 	Number of reports prescribed the drug that had the condition
        B 	integer 	Number of reports prescribed the drug that did not have the condition
        C 	integer 	Number of reports not prescribed the drug† that had the condition
        D 	integer 	Number of reports not prescribed the drug† that did not have the condition
        PRR 	float 	Proportional reporting ratio*
        PRR_error 	float 	Proportional reporting ratio error*
        mean_reporting_frequency 	float 	a / (a + b)

        Further documentation see here:
        https://github.com/tatonetti-lab/nsides-release/blob/master/release-notes/v0.1.md
        """
        file_path = get_file_path(self.urls[OFFSIDES], self.biodb_name)
        df = pd.read_csv(file_path, low_memory=False)
        df.drop_duplicates(inplace=True)
        df["source"] = "offsides"
        df.columns = self._standardize_column_names(df.columns)

        df.index += 1
        df.index.rename("id", inplace=True)
        df[pd.to_numeric(df["condition_meddra_id"], errors="coerce").notnull()].to_sql(
            nsides.Nsides.__tablename__,
            self.engine,
            if_exists="append",
            chunksize=100000,
        )
        return {self.biodb_name: df.shape[0]}

    # TODO: Reimplement, but structure have changed
    # def __import_onsides(self) -> Optional[pd.DataFrame]:
    #     """Extract OnSIDES CSV file and format into a DF."""
    #     file_path = get_file_path(self.urls[ONSIDES], self.biodb_name)
    #     file_folder = os.path.dirname(file_path)
    #     tar = tarfile.open(file_path, "r:gz")
    #     folder_in_tar_file = tar.members[0].path
    #     fd = tar.extractfile(f"{folder_in_tar_file}/adverse_reactions.csv.gz")
    #     df = pd.read_csv(fd, compression="gzip", encoding="utf-8")

    #     df.drop(  # Remove columns that aren't needed
    #         [
    #             "xml_id",
    #             "Unnamed: 10",
    #             "omop_concept_id",
    #             "drug_concept_ids",
    #             "concept_class_id",
    #         ],
    #         inplace=True,
    #         axis=1,
    #     )

    #     df.rename(
    #         columns={  # Rename columns to match OFFSIDES
    #             "rxnorm_ids": "drug_rxnorn_id",
    #             "pt_meddra_term": "condition_concept_name",
    #             "pt_meddra_id": "condition_meddra_id",
    #             "ingredients_names": "drug_concept_name",
    #         },
    #         inplace=True,
    #     )

    #     # Keep rows with only 1 ingredient/drug
    #     single_value_mask = df["drug_concept_name"].apply(
    #         lambda x: len(x.split(",")) == 1
    #     )
    #     pruned_df = df.loc[single_value_mask]
    #     pruned_df["source"] = "onsides"
    #     return pruned_df

    def update_bel(self) -> int:
        """Create has_side_effect edges between drugs (drugbank) and side_effects.

        RxCUI (https://rxnav.nlm.nih.gov/RxNormAPIs.html#) is used as the mapping identifier.
        """
        self.clear_edges()
        self.delete_nodes_with_no_edges("side_effect")
        self.delete_nodes_with_no_edges("drug")

        # TODO: Translate to sqlalchemy query
        sql_temp = """Select
            o.condition_meddra_id,
            o.condition_concept_name,
            o.prr,
            o.mean_reporting_frequency
        from
            drugbank as d inner join
            drugbank_external_identifier as dei on (d.id=dei.drugbank_id) inner join
            nsides as o on (dei.identifier=o.drug_rxnorn_id)
        where
            d.drugbank_id='{}' and resource='RxCUI'
            and (mean_reporting_frequency>=0.01 OR mean_reporting_frequency is NULL)
        group by
            o.condition_meddra_id,
            o.condition_concept_name,
            o.prr,
            o.mean_reporting_frequency
        """

        drugbank_ids = self.query_class("drug", columns=["drugbank_id"], drugbank_id="notnull")
        drugbank_id_rids = {d["drugbank_id"]: d[RID] for d in drugbank_ids}

        side_effects = self.query_class("side_effect", columns=["condition_meddra_id"])
        side_effect_rids = {d["condition_meddra_id"]: d[RID] for d in side_effects}

        updated = 0

        for drugbank_id, drugbank_rid in tqdm(drugbank_id_rids.items(), desc=f"Update {self.biodb_name.upper()}"):
            for r in self.engine.execute(sql_temp.format(drugbank_id)):
                (
                    condition_meddra_id,
                    condition_concept_name,
                    prr,
                    mean_reporting_frequency,
                ) = r

                if condition_meddra_id not in side_effect_rids:
                    side_effect_rid = self.insert_record(
                        "side_effect",
                        {
                            "label": condition_concept_name,
                            "condition_meddra_id": condition_meddra_id,
                        },
                    )
                    side_effect_rids[condition_meddra_id] = side_effect_rid

                value_dict = {
                    "prr": prr,
                    "mean_reporting_frequency": mean_reporting_frequency,
                }
                side_effect_rid = side_effect_rids[condition_meddra_id]
                self.create_edge(
                    class_name="has_side_effect",
                    from_rid=drugbank_rid,
                    to_rid=side_effect_rid,
                    value_dict=value_dict,
                )
                updated += 1
        return updated

    def update_interactions(self) -> int:
        """Update edges with NSIDES metadata."""
        # not implemented because update_bel is implemented
        pass



"""KEGG API methods."""

from flask.globals import request
from sqlalchemy import or_

from ebel.manager.rdbms.models.kegg import Kegg
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_data, _get_paginated_ebel_query_result,
                                  _get_paginated_query_result)


def get_kegg():
    """Get generic KEGG entry."""
    return _get_data(Kegg)


def get_by_gene_symbol():
    """Get KEGG entry by gene symbol."""
    symbol = request.args.get("gene_symbol")
    query_filter = or_(Kegg.gene_symbol_a == symbol, Kegg.gene_symbol_b == symbol)
    query = RDBMS.get_session().query(Kegg).filter(query_filter)
    return _get_paginated_query_result(query)


def get_ebel_relation():
    """Get KEGG related eBEL relations."""
    conf = {
        "interaction_type": "interaction_type",
        "pathway_names": "pathway_name",
        "edge_id": "@rid.asString()",
        "interactor_a_rid": "out.@rid.asString()",
        "interactor_a_name": "out.name",
        "interactor_a_namespace": "out.namespace",
        "interactor_a_bel": "out.bel",
        "interactor_b_rid": "in.@rid.asString()",
        "interactor_b_namespace": "in.namespace",
        "interactor_b_name": "in.name",
        "interactor_b_bel": "in.bel",
        "relation": "@class.asString()",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_ppi_kg"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    pathway_name = request.args.get("pathway_name")
    if paras or pathway_name:
        wheres = []
        if paras:
            wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        if pathway_name:
            wheres += [f'"{pathway_name}" in pathway_name']

        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""Protein Atlas API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestProteinAtlas:
    def test_get_rna_brain_fantom(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/rna_brain_fantom?gene=ENSG00000105383&gene_name=CD33&brain_region=amygdala&tags_per_million=10.7&scaled_tags_per_million=12.4&nx=3.6&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "brain_region": "amygdala",
            "gene": "ENSG00000105383",
            "gene_name": "CD33",
            "nx": "3.6",
            "scaled_tags_per_million": "12.4",
            "tags_per_million": "10.7",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_rna_mouse_brain_allen(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/rna_mouse_brain_allen?gene=ENSG00000095970&gene_name=TREM2&brain_region=amygdala&expression_energy=0.5&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "brain_region": "amygdala",
            "expression_energy": 0.5,
            "gene": "ENSG00000095970",
            "gene_name": "TREM2",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_normal_tissue(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/normal_tissue?gene=ENSG00000105383&gene_name=CD33&tissue=adipose%20tissue&cell_type=adipocytes&level=Not%20detected&reliability=Approved&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "cell_type": "adipocytes",
            "gene": "ENSG00000105383",
            "gene_name": "CD33",
            "level": "Not detected",
            "reliability": "Approved",
            "tissue": "adipose tissue",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_subcellular_location(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/rna_subcellular_location?gene=ENSG00000105383&gene_name=CD33&reliability=Approved&main_location=Nucleoplasm%3BPlasma%20membrane&supported=Plasma%20membrane&uncertain=Nucleoplasm&go_id=Nucleoplasm%20%28GO%3A0005654%29%3BPlasma%20membrane%20%28GO%3A0005886%29&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "additional_location": None,
            "approved": None,
            "cell_cycle_dependency": None,
            "enhanced": None,
            "extracellular_location": None,
            "gene": "ENSG00000105383",
            "gene_name": "CD33",
            "go_id": "Nucleoplasm (GO:0005654);Plasma membrane (GO:0005886)",
            "main_location": "Nucleoplasm;Plasma membrane",
            "reliability": "Approved",
            "single_cell_variation_intensity": None,
            "single_cell_variation_spatial": None,
            "supported": "Plasma membrane",
            "uncertain": "Nucleoplasm",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_rna_tissue_consensus(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/rna_tissue_consensus?gene=ENSG00000105383&gene_name=CD33&tissue=adipose%20tissue&nx=10.2&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {"gene": "ENSG00000105383", "gene_name": "CD33", "nx": 10.2, "tissue": "adipose tissue"}
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_rna_brain_gtex(self, client):
        response = client.get(
            "api/v1/ebel/protein_atlas/rna_brain_gtex?gene=ENSG00000000003&gene_name=TSPAN6&brain_region=amygdala&tpm=7.3&p_tpm=9.0&nx=7.0&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "brain_region": "amygdala",
            "gene": "ENSG00000000003",
            "gene_name": "TSPAN6",
            "nx": 7,
            "p_tpm": 9,
            "tpm": 7.3,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""IUPHAR drug module."""

import logging
from typing import Dict

import numpy as np
import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.uniprot import UniProt
from ebel.manager.orientdb.constants import IUPHAR
from ebel.manager.rdbms.models import iuphar
from ebel.tools import get_file_path

logger = logging.getLogger(__name__)


class Iuphar(odb_meta.Graph):
    """IUPHAR."""

    def __init__(self, client: OrientDB = None):
        """Init IUPHAR."""
        self.client = client
        self.biodb_name = IUPHAR
        self.urls = {
            "iuphar_int": urls.IUPHAR_INT,
            "iuphar_ligands": urls.IUPHAR_LIGANDS,
        }
        super().__init__(
            edges=odb_structure.iuphar_edges,
            tables_base=iuphar.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        ligands = self.insert_ligand_data()
        ints = self.insert_interaction_data()

        self.session.commit()
        return {"ligands": ligands, "interactions": ints}

    def insert_ligand_data(self) -> int:
        """Insert ligand/drug data in generic OrientDB class."""
        logger.info("Reading and formatting IUPHAR drug data")
        df = pd.read_csv(
            get_file_path(self.urls["iuphar_ligands"], self.biodb_name),
            sep=",",
            low_memory=False,
            skiprows=[0],
            # dtype={'Ligand id': 'Int64', 'PubChem SID': 'Int64', 'PubChem CID': 'Int64'},
            true_values=["yes"],
        ).replace(
            {np.nan: None}
        )  # Convert 'yes' to True
        df.columns = self._standardize_column_names(df.columns)
        df.rename(
            columns={
                "in_ch_i": "inchi",
                "in_ch_ikey": "inchi_key",
                "pub_chem_cid": "pubchem_cid",
                "pub_chem_sid": "pubchem_sid",
                "uni_prot_id": "uniprot_id",
                "ligand_id": "id",
            },
            inplace=True,
        )  # Make it consistent with drugbank
        df.set_index("id", inplace=True)
        df.to_sql(iuphar.IupharLigand.__tablename__, self.engine, if_exists="append")

        return df.shape[0]

    def insert_interaction_data(self) -> int:
        """Insert interaction data in generic OrientDB class."""
        logger.info("Reading and formatting IUPHAR interaction data.")
        df = pd.read_csv(
            get_file_path(self.urls["iuphar_int"], self.biodb_name),
            sep=",",
            low_memory=False,
            skiprows=[0],
            dtype={
                "target_id": "Int64",
                "ligand_pubchem_sid": "Int64",
                "ligand_id": "Int64",
                "target_ligand_id": "Int64",
                "target_ligand_pub_chem_sid": "Int64",
            },
            true_values=["t"],
            false_values=["f"],
        ).replace({np.nan: None})

        df.columns = self._standardize_column_names(df.columns)
        df.rename(
            columns={
                "target_uni_prot_id": "target_uniprot",
                "target_ligand_uni_prot_id": "target_ligand_uniprot_id",
                "target_ligand_pub_chem_sid": "target_ligand_pubchem_sid",
                "ligand_pub_chem_sid": "ligand_pubchem_sid",
                "pub_med_id": "pubmed_id",
            },
            inplace=True,
        )
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(iuphar.IupharInteraction.__tablename__, self.engine, if_exists="append")

        return df.shape[0]

    def update_interactions(self) -> int:
        """Update the ligands/drug generic table to include a link to all interactions it is part of."""
        self.clear_edges()
        uniprot = UniProt(self.client)

        iuphar_edge_type_mapper = {
            "Agonist": "agonist_of__iu",
            "Inhibitor": "inhibits__iu",
            "Antagonist": "antagonist_of__iu",
            "Channel blocker": "channel_blocker_of__iu",
            "Allosteric modulator": "allosteric_modulator_of__iu",
            "Activator": "activates__iu",
            "Antibody": "antibody_against__iu",
            "Gating inhibitor": "inhibits_gating__iu",
        }

        sql = """select i.pubmed_id, i.assay_description, i.affinity_units, i.affinity_low, i.affinity_median,
        i.affinity_high, i.type,
        i.action,i.target_uniprot, l.name as ligand_name, l.pubchem_sid, i.ligand_gene_symbol, i.ligand_species
        from iuphar_interaction as i inner join iuphar_ligand as l
        on (i.ligand_id=l.id) where i.target_uniprot IS NOT NULL and pubchem_sid IS NOT NULL"""

        df_iuphar = pd.read_sql(sql, self.engine).replace({np.nan: None})
        df_iuphar.set_index("target_uniprot", inplace=True)
        df_graph = pd.DataFrame(
            uniprot.get_pure_uniprot_rid_dict_in_bel_context().items(),
            columns=["target_uniprot", "rid"],
        )
        df_graph.set_index("target_uniprot", inplace=True)
        df_join = df_graph.join(df_iuphar, how="inner")

        for uniprot, data in tqdm(
            df_join.iterrows(),
            total=df_join.shape[0],
            desc=f"Update {self.biodb_name.upper()} interactions",
        ):
            if data.ligand_gene_symbol and data.ligand_species and "Human" in data.ligand_species:
                symbol = data.ligand_gene_symbol.split("|")[0]  # human seems to always the first
                a_value_dict = {
                    "pure": True,
                    "bel": f'p(HGNC:"{symbol}")',
                    "namespace": "HGNC",
                    "name": symbol,
                }
                a_class = "protein"
            else:
                a_value_dict = {
                    "pure": True,
                    "bel": f'a(PUBCHEM:"{data.pubchem_sid}")',
                    "namespace": "PUBCHEM",
                    "name": str(data.pubchem_sid),
                    "label": data.ligand_name,
                }
                a_class = "abundance"
            a_rid = self.get_create_rid(a_class, value_dict=a_value_dict, check_for="bel")

            i_value_dict = {
                "pmids": data.pubmed_id.split("|") if data.pubmed_id else None,
                "assay_description": data.assay_description,
                "affinity_units": data.affinity_units,
                "affinity_low": data.affinity_low,
                "affinity_median": data.affinity_median,
                "affinity_high": data.affinity_high,
                "type": data.type,
                "action": data.action,
            }
            edge_class = iuphar_edge_type_mapper.get(data.type, "iuphar_interaction")
            self.create_edge(edge_class, from_rid=a_rid, to_rid=data.rid, value_dict=i_value_dict)

        # not sure if this is really needed
        # Hgnc(self.client).update_bel()

        return df_join.shape[0]



"""KEGG RDBMS model definition."""
from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Kegg(Base):
    """Class definition for the kegg table."""

    __tablename__ = "kegg"
    id = Column(Integer, primary_key=True)

    pathway_identifier = Column(String(100))
    pathway_name = Column(String(1000))
    kegg_species_id = Column(String(100))
    kegg_gene_id_a = Column(String(100))
    gene_symbol_a = Column(String(100), index=True)
    kegg_gene_id_b = Column(String(100))
    gene_symbol_b = Column(String(100), index=True)
    kegg_int_type = Column(String(100))
    interaction_type = Column(String(50), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""UniProt API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestUniprot:
    def test_get_uniprot(self, client):
        response = client.get(
            "api/v1/uniprot?accession=P20138&gene_symbol=CD33&taxonomy_id=9606&keyword=Phosphoprotein&xref_db=ChEMBL&xref_id=CHEMBL1842&subcellular_location=Peroxisome&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_sample = {  # Only check a subset of data
            "accession": "P20138",
            "gene_names": ["CD33", "SIGLEC3"],
            "gene_symbol": "CD33",
            "taxid": 9606,
            "name": "CD33_HUMAN",
            "organism": "Homo sapiens",
            "recommended_name": "Myeloid cell surface antigen CD33",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([hit[key] == value for key, value in expected_sample.items()])

    def test_get_keyword_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/keyword/starts_with?keyword=Phago&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "Phagocytosis" in results
        assert isinstance(results["Phagocytosis"], int)

    def test_get_subcellular_location_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/subcellular_location/starts_with?subcellular_location=Endo&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 10
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "Endoplasmic reticulum-Golgi intermediate compartment" in results
        assert isinstance(results["Endoplasmic reticulum-Golgi intermediate compartment"], int)

    def test_get_gene_symbol_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/gene_symbol/starts_with?gene_symbol=CD18&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {"CD180": 2690, "Cd180": 22498}
        assert output[NUM_RESULTS] == 2
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results

    def test_get_gene_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/gene/starts_with?gene=CD11&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_results = {"CD1176A": 294700, "CD11A": 1040093, "CD11B": 1040099, "CD11C": 1040111}
        assert output[NUM_RESULTS] == 4
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results

    def test_get_organism_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/organism/starts_with?organism=Homo%20sapiens&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {"Homo sapiens": 9606, "Homo sapiens neanderthalensis": 63221}
        assert output[NUM_RESULTS] == 2
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results

    def test_get_function_starts_with(self, client):
        response = client.get(
            "api/v1/uniprot/function/starts_with?description=Phosphatase%202A&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "Phosphatase 2A affects a variety of biological processes in the cell such as transcription, cell cycle progression and cellular morphogenesis, and provides an initial identification of critical substrates for this phosphatase. The regulatory subunit may direct the catalytic subunit to distinct, albeit overlapping, subsets of substrates (By similarity).": 15,
            "Phosphatase 2A affects a variety of biological processes in the cell such as transcription, cell cycle progression and cellular morphogenesis, and provides an initial identification of critical substrates for this phosphatase. The regulatory subunit may direct the catalytic subunit to distinct, albeit overlapping, subsets of substrates.": 16,
        }
        assert output[NUM_RESULTS] == 2
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results



"""Error class definitions."""
import re
import sys
from collections import OrderedDict

from lark.exceptions import UnexpectedCharacters, UnexpectedToken

TEMPLATE = (
    "{error_class}\tkeyword:{keyword}\tentry:{entry}\tline:{line_number}\tcolumn:{column}" "\turl:{url}\thint:{hint}"
)


class _Error:
    """Base class for all errors."""

    def __init__(self):
        self.class_name = self.__class__.__name__
        self.value_dict = OrderedDict(
            [
                ("error_class", self.class_name),
                ("url", None),
                ("keyword", None),
                ("entry", None),
                ("line_number", None),
                ("column", None),
                ("hint", None),
            ]
        )

    def to_dict(self):
        """Format the properties of error into a dictionary."""
        raise NotImplementedError("to_dict have to be implemented in {}".format(self.__class__.__name__))

    def to_string(self) -> str:
        """Format the output to a string."""
        self.value_dict.update(self.to_dict())
        return TEMPLATE.format(**self.value_dict)

    def __str__(self):
        return self.to_string()


class NotInNamespacePattern(_Error):
    """Error in entry link to a namespace defined as pattern in the header but does not fit the pattern."""

    def __init__(self, keyword: str, entry: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(NotInNamespacePattern, self).__init__()

        self.ns_keyword = keyword
        self.entry = entry
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class NotInAnnotationPattern(_Error):
    """Error in entry links to a namespace defined as patter in the header but not fits the pattern."""

    def __init__(self, keyword: str, entry: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(NotInAnnotationPattern, self).__init__()

        self.ns_keyword = keyword
        self.entry = entry
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class WithoutDefinedAnnotation(_Error):
    """Error in entry without a defined annotation."""

    def __init__(self, entry: str, keyword: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(WithoutDefinedAnnotation, self).__init__()

        self.entry = entry
        self.keyword = keyword
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class WithoutDefinedNamespace(_Error):
    """Error in entry without a defined namespace."""

    def __init__(self, entry: str, keyword: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(WithoutDefinedNamespace, self).__init__()

        self.entry = entry
        self.keyword = keyword
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class NotInNamespaceUrl(_Error):
    """Error in entry links to a namespace defined in the header but does not exist in namespace url."""

    def __init__(
        self,
        keyword: str,
        url_or_path: str,
        entry: str,
        line_number: int,
        column: int,
        hint: str,
    ):
        """Initialize error class.

        :param keyword: Error class type.
        :param url_or_path: URL or directory path to namespace or annotation file.
        :param hint: Sentence or string to give the user in the error report to help solve error.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        :param entry: String representation where error occurred.
        """
        super(NotInNamespaceUrl, self).__init__()

        self.ns_keyword = keyword
        self.url_or_path = url_or_path
        self.entry = entry
        self.line_number = line_number
        self.column = column
        self.hint = hint

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "url": self.url_or_path,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
            "hint": self.hint,
        }


class NotInAnnotationUrl(_Error):
    """Error in entry links to an annotation defined in the header but not exists in namespace url."""

    def __init__(
        self,
        keyword: str,
        url_or_path: str,
        entry: str,
        line_number: int,
        column: int,
        hint: str,
    ):
        """Initialize error class.

        :param keyword: Error class type.
        :param url_or_path: URL or directory path to namespace or annotation file.
        :param hint: Sentence or string to give the user in the error report to help solve error.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        :param entry: String representation where error occurred.
        """
        super(NotInAnnotationUrl, self).__init__()

        self.ns_keyword = keyword
        self.url_or_path = url_or_path
        self.entry = entry
        self.line_number = line_number
        self.column = column
        self.hint = hint

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "url": self.url_or_path,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
            "hint": self.hint,
        }


class NotInNamespaceList(_Error):
    """Error in entry links to a namespace defined in the header but not exists in namespace list."""

    def __init__(self, keyword: str, entry: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(NotInNamespaceList, self).__init__()

        self.ns_keyword = keyword
        self.entry = entry
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class NotInAnnotationList(_Error):
    """Entry links to an annotation defined in the header but does not exist in the annotation list."""

    def __init__(self, keyword: str, entry: str, line_number: int, column: int):
        """Initialize error class.

        :param keyword: Error class type.
        :param entry: String representation where error occurred.
        :param line_number: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(NotInAnnotationList, self).__init__()

        self.ns_keyword = keyword
        self.entry = entry
        self.line_number = line_number
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "keyword": self.ns_keyword,
            "entry": re.sub("[\n\r]", "", self.entry),
            "line_number": self.line_number,
            "column": self.column,
        }


class BelSyntaxError(_Error):
    """Syntax error."""

    def __init__(self, exception, line_number: int, line: str):
        """Initialize error class.

        :param exception: Error class type.
        :param line_number: URL or directory path to namespace or annotation file.
        :param line: Line where error occurred.
        """
        super(BelSyntaxError, self).__init__()

        self.line_number = line_number
        self.line = line
        self.exception = exception

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        hint = (
            "%s >>>>>> %s"
            % (
                self.line[: self.exception.column - 1],
                self.line[self.exception.column - 1 :],
            )
        ).strip()

        value_dict = {
            "line_number": self.line_number,
            "hint": re.sub("[\n\r]", "", hint),
        }

        if isinstance(self.exception, UnexpectedCharacters):
            offset = self.exception.column
            value_dict.update(
                {
                    "error_class": self.__class__.__name__ + "_unexpected_input",
                    "entry": self.line[offset : (offset + 5)],
                    "column": self.exception.column,
                }
            )

        elif isinstance(self.exception, UnexpectedToken):
            value_dict.update(
                {
                    "error_class": self.__class__.__name__ + "_unexpected_token",
                    "entry": re.sub("[\n\r]", "", self.exception.token),
                    "column": self.exception.column,
                }
            )
        else:
            print(
                "Not covered by library: lark.expections type {}".format(type(self.exception)),
                sys.exc_info()[0],
            )
            raise

        return value_dict


class NotDownloadedFromUrl(_Error):
    """Error in entry links to an annotation defined in the header but not exists in namespace url."""

    def __init__(
        self,
        keyword: str,
        url_or_path: str,
        hint: str,
        line: int = None,
        column: int = None,
    ):
        """Initialize error class.

        :param keyword: Error class type.
        :param url_or_path: URL or directory path to namespace or annotation file.
        :param hint: Sentence or string to give the user in the error report to help solve error.
        :param line: Line number where error was found.
        :param column: Column number where error was found.
        """
        super(NotDownloadedFromUrl, self).__init__()

        self.ns_keyword = keyword
        self.url_or_path = url_or_path
        self.hint = hint
        self.line = line
        self.column = column

    def to_dict(self) -> dict:
        """Format the properties of error into a dictionary."""
        return {
            "error_class": self.class_name,
            "url": self.url_or_path,
            "keyword": self.ns_keyword,
            "column": self.column,
            "line_number": self.line,
            "hint": self.hint,
        }
